; /////////////////////////////////////////////////////////////////////////////////
;//                                                                               //
;//  This program and its associated documentation are provided for your personal //
;//  use only and appear here exclusively by permission of the copyright holder.  //
;//  Please contact the copyright holder before re-distributing, re-publishing    //
;//  or disseminating this copyrighted work. This code is not GPL or in the       // 
;//  public domain. Please respect the author's copyright.                        //
;//                                                                               //
;//  No waranty, either expressed or implied, are given.  I assume no liability   //
;//  for its use in any project or device.                                        //
;//                                                                               //
;//  Your use of this program indicates your acceptance of all license terms.     //
;//  This particular version is freeware as long as the copyright messages are    //
;//  left intact.                                                                 //  
;//                                                                               //
; /////////////////////////////////////////////////////////////////////////////////
;
; SBC-3 Keyboard and RS-232 Interface Firmware
; 
; SPI interface to RS-232 and PC Keyboard ports
;
; (c)2008 Daryl Rictor 
;
; V1.0    11-03-2008
;
;
.include  "m8def.inc"				; standard definitions
.include  "regdef.inc"				; Application definitions

;******************** Info **********************
;
; SPI Commands
;
; $00 - null (used to receive data)
; $01 - GET STATUS      
;	bit 7  KB data avail
;       bit 6  RX data avail
;       bit 5  TX ready
;       bit 4  RX Frame error 
;       bit 3  RX overrun
;       bit 2  RX parity error
;       bit 1  KB missing
;       bit 0  not used
;
; $02 - get RX byte
; $03 - send TX byte
; $04 - get ser parms
; $05 - set ser parms : bit 0-2 = speed, bit 3 = 7/8 bit, bit 4 = parity y/n, bit 5 = parity odd/even, bit 6 = stop 1/2
; $06 - get KB byte 
; $07 - send kb cmd   0= ASCII, 1= scancode 2= future > 2 = KB CMD byte
; $08 - init RS232
; $09 - init KB


;******************** ATM8 Port Usage **********************
;
; Port B
;
; 0 not used - pad
; 1 not used - pad
; 2 SS
; 3 MOSI
; 4 MISO
; 5 SCLK
; 6 not used
; 7 not used
;
;Port C
;
; 0 KBData
; 1 not used
; 2 not used
; 3 not used (Alt-Ctrl-Del support)
; 4 not used
; 5 IRQ out
; 6 /RESET in
; 7 n/a

.set  kbdat     	= PC0			; bi directional
.set  IRQ               = PC5                   ; IRQ to CB2 of VIA1

;Port D 
;
; 0 RS-232 RX
; 1 RS-232 TX
; 2 not used (INT0)
; 3 KBCLK (INT1)
; 4 RS-232 CLK in (not implemented yet)
; 5 not used
; 6 not used
; 7 not used
;
.set  kbclk     	= PD3			; bi directional INT1
;

;******************************************************************************
; Code Begins at 0x000
; System IRQ jump table from 0x0000 - 0x0012
;******************************************************************************
	rjmp 	RESET 				; Reset Handler
	rjmp	EXT_INT0 			; IRQ0 Handler
	rjmp 	EXT_INT1 			; IRQ1 Handler
	rjmp 	TIM2_COMP 			; Timer2 Compare Handler
	rjmp 	TIM2_OVF 			; Timer2 Overflow Handler
	rjmp 	TIM1_CAPT 			; Timer1 Capture Handler
	rjmp 	TIM1_COMPA 			; Timer1 CompareA Handler
	rjmp 	TIM1_COMPB 			; Timer1 CompareB Handler
	rjmp 	TIM1_OVF 			; Timer1 Overflow Handler
	rjmp 	TIM0_OVF 			; Timer0 Overflow Handler
	rjmp 	SPI_STC 			; SPI Transfer Complete Handler
	rjmp 	USART_RXC 			; USART RX Complete Handler
	rjmp 	USART_UDRE 			; UDR Empty Handler
	rjmp 	USART_TXC	 		; USART TX Complete Handler
	rjmp 	ADCC 				; ADC Conversion Complete Handler
	rjmp 	EE_RDY	 			; EEPROM Ready Handler
	rjmp 	ANA_COMP 			; Analog Comparator Handler
	rjmp 	TWSI 				; Two-wire Serial Interface Handler
	rjmp 	SPM_RDY 			; Store Program Memory Ready Handler
	
;******************************************************************************
; unused IRQ vector's point to software reset code
;******************************************************************************

EXT_INT0:
;EXT_INT1:   ; used in keyboard to detect clk pin changes
TIM2_COMP:
TIM2_OVF:
TIM1_CAPT:
TIM1_COMPA:
TIM1_COMPB:
;TIM1_OVF:   ; used in keyboard to prevent lockup on no kb attached
TIM0_OVF:
SPI_STC:
USART_RXC:
USART_UDRE:
USART_TXC:
ADCC:
EE_RDY:
ANA_COMP:
TWSI:
SPM_RDY:

;******************************************************************************
; Program Starts here on Reset
;******************************************************************************
RESET:						; first line executed after reset 
	CLI					; disable interupts
	ldi	i, 0x04				; set stack pointer to top of SRAM 
	out	SPH, i				; 
	ldi	i, 0x5f				; 
	out	SPL, i				;

;******************** OSCCAL data is chip specific!! **********************
	ldi	ZH, 0x1F			; last word in Flash
	ldi	ZL, 0xFE			;  
	lpm	i, Z+				; get cal byte for 8MHz
	lpm	j, Z				; get byte valid flag
						; if bit 7 of last byte is 1, then valid osccal
	sbrc	j, 7				; if its 0, then don't store value 
	out	OSCCAL, i			; if its 1, store it
;**************************************************************************
; port B pins (SPI I/O)
	LDI	i, 0x10				; all inputs, except MISO
	out	ddrb, i				; 
	LDI	i, 0xC4				; pull-ups on unused pins and SS
	out	portb, i			; 

; port C pins (KB & IRQ I/O)
	LDI	i, 0x20				; IRQ output pin 5, all others inputs
	out	ddrc, i				; 
	LDI	i, 0x3E				; pull-ups on unused pins, IRQ = 1, res no pull-up
	out	portc, i			; 

; port D pins (RS232 & INT1 I/O)
	LDI	i, 0xF4				; pullups on all unused pins and INT0 and RS-232 CLK in
	out	portd, i			; 
	LDI	i, 0x08				; all inputs except PD2 KBCLK held low
	out	ddrd, i				; 

;**************************************************************************
; set INT1 IRQ for kb clk
	ldi	i, 0x08				; FALLING EDGE int1 SENSE
	out	MCUCR, i	
	ldi	i, 0x00
	out	GICR, i				; diable INT1
	ldi	i, 0x80
	out	GIFR, i				; clear active INT1 flags

;**************************************************************************
	rcall	SPI_init			; init SPI interface
	rcall	USART_Init			; init USART
	rcall	KB_Init				; Init keyboard decoder
	SEI					; Ebable interrupts
NOKB:
	cbi	portc, IRQ			; send ready flag
	nop					; delay
	sbi	portc, IRQ			; restore flag

;******************************************************************************
; Main Loop
;******************************************************************************
MAIN:
	rcall	SPI_RX				; process any data from SPI master
	rcall	USART_ST			; update USART status
	sbic	pinc, IRQ			; if IRQ, wait for SPI to finish
	rcall	USART_RX			; store RS-232 rx data in FIFO
	sbic	pinc, IRQ			; if IRQ, wait for SPI to finish
	rcall	KBDATRDY			; store KB data in FIFO
	rjmp	main				; repeat

;******************************************************************************
; SPI Routines 
;******************************************************************************

SPI_Init:					; Initialize SPI interface
	ldi	i, 0x40				;
	out	SPCR, i				; enable spi, slave mode
	in	i, SPSR				;
	in	i, SPDR				; clr status reg
	ldi	SPIST, 0x00			; clear system status reg
	ldi	SMODE, 0x00			; 
	ret

SPI_RX:     					; If SPI data available, process it
	sbis 	SPSR,SPIF			;
	ret					; nothing received
	in	i, SPSR				; clear flags
	in	i, SPDR				; get received byte from SPI
	tst	SMODE				; SPI Modes  $00=Command, $01=RS232 TX, $02=RS232 PARM, $03=KB CMD
	brne	SPIDATP				;
	cpi	i, 0x00				;
	brne	SPI_RX1				;
	sbi	portc, IRQ			;
	ret					;
SPI_RX1:
	cpi	i, 0x01				;
	brne	SPI_RX2				;
	out	SPDR, SPIST			;
	cbi	portc, IRQ			;
	ret					;
SPI_RX2:
	cpi	i, 0x02				;
	brne	SPI_RX3				;
	rjmp	USART_RX_SPI			;
SPI_RX3:
	cpi	i, 0x03				;
	brne	SPI_RX4				;
	ldi	SMODE, 0x01			;
	cbi	portc, IRQ			;
	ret					;
SPI_RX4:
	cpi	i, 0x04				;
	brne	SPI_RX5				;
	rjmp	USART_GPARM			;
SPI_RX5:
	cpi	i, 0x05				;
	brne	SPI_RX6				;
	ldi	SMODE, 0x02			;
	cbi	portc, IRQ			;
	ret					;
SPI_RX6:
	cpi	i, 0x06				;
	brne	SPI_RX7				;
	sbrs    SPIST, 1			; if kb missing, don't try to read it
	rjmp	KB_RX_SPI			;
	ret					;
SPI_RX7:
	cpi	i, 0x07				;
	brne	SPI_RX8				;
	ldi	SMODE, 0x03			;
	cbi	portc, IRQ			;
	ret					;
SPI_RX8:
	cpi	i, 0x08				;
	brne	SPI_RX9				;
	rjmp	USART_INIT			;
SPI_RX9:
	cpi	i, 0x09				;
	brne	SPI_RX10			;
	rjmp	KB_INIT				;
SPI_RX10:
	ret					; invalid command, ignore

SPIDATP:					; parse data mode type
	sbi	portc, IRQ			;
	cpi	SMODE, 0x01			;
	breq	SENDSER				;
	cpi	SMODE, 0x02			;
	breq	SENDPM				;
	cpi	SMODE, 0x03			;
	breq	SENDKB				;
	clr	SMODE				; invalid SMODE, ignore
	ret					;

SENDSER:
	clr	SMODE  				;
	rjmp	USART_TX			;

SENDPM:
	clr	SMODE  				;
	rjmp	USART_SPARM			;

SENDKB:
	clr	SMODE  				;
	sbrc    SPIST, 1			; if kb missing, don't send
	ret					;
	cpi	i, 0x00				;
	breq	SENDKB1				;
	cpi	i, 0x01				;
	breq	SENDKB2				;
	rjmp	SEND				;

SENDKB1: 
	ldi	KBMODE, 0x00			; ASCII mode
	ret				
SENDKB2:
	ldi	KBMODE, 0x80			; SCANCODE mode
	ret
;
;******************************************************************************
;  USART routines 
;******************************************************************************
;
USART_Init:
	ldi	i, 0x0A				; 9600,N,8,1
	rjmp	USART_SPARM

USART_RX:
	sbis	UCSRA, RXC  			; nothing to receive
	ret
	inc	RBST
	breq	USART_RX1			; buffer full 
	ori	SPIST, 0x40			; RX data waiting = 1
	movw	YL, RBIN
	in	i, UDR
	st	Y+, i
	mov	RBIN, YL
	ret
USART_RX1:
	in	i, UDR				; flush new data
	dec	RBST				; mark buffer full
	ret

USART_RX_SPI:  ; move rs-232 rx data to spi port
	movw	YL, RBOUT
	ld	i, Y+
	out	SPDR, i				; send to SPI 
	mov	RBOUT, YL
	dec	RBST
	brne	USART_RX_SPI1
	andi	SPIST, 0xBF			; RX data rdy=0 
USART_RX_SPI1:
	cbi	portc, IRQ			;
	ret

USART_TX:					; send byte in i
  	sbis	UCSRA,UDRE			; Wait for empty transmit buffer
	rjmp	USART_TX
	out	UDR, i
	ret

USART_ST:					; update SPIST with USART bits 
	in	i, UCSRA
	andi	i, 0x3C
	andi	SPIST, 0xC3
	or	SPIST, i
	ret

USART_GPARM:
	out	SPDR, SERP				; send to SPI 
	cbi	portc, IRQ			;
        ret

USART_SPARM:
; bit 0-2 = speed, bit 3 = 7/8 bit, bit 4 = parity odd/even, bit 5 = parity y/n, bit 6 = stop 1/2
;
;       0 = 2400  = 207 = CF      0 = 7 bit      00 = no parity       0 = 1 stop bits
;       1 = 4800  = 103 = 67      1 = 8 bit      01 = no parity       1 = 2 stop bits
;       2 = 9600  = 51  = 33                     10 = even parity
;       3 = 19200 = 25  = 19                     11 = odd parity
;     4-7 = 38400 = 12  = 0C
;        
; default mode is 9600, 8N1, or SERP = 00001010 - $0A    

        mov	SERP, i
	ldi	j, 0x00
	out	UCSRB, j			; disable Receiver and Transmitter

	ldi	j, 0xCF  ; 2400
        andi    i, 0x07
	cpi 	i, 0x00
        breq    USPARM5
	ldi	j, 0x67 ; 4800
	cpi	i, 0x01
	breq	USPARM5
	ldi	j, 0x33 ; 9600
	cpi	i, 0x02
	breq	USPARM5
	ldi	j, 0x19 ; 19200
	cpi	i, 0x03
	breq	USPARM5
	ldi	j, 0x0C ; 38400        
USPARM5: 
	out	UBRRL, j
	ldi	i, 0x00
	out	UBRRH, i
	
	mov	i, SERP
	ldi	j, 0x80				; enable write
	sbrc    i, 3				; bit 3
	ori	j, 0x06				;
	sbrc    i, 5				; bit 5
	rjmp	USPARM6				;
	sbrc    i, 4				; bit 4
	ori	j, 0x10				;
USPARM6: 
	sbrc    i, 6				; bit 6
        ori	j, 0x08				; 
	out	UCSRC, j			; set parity, word len and stop bits

	ldi	i, 0x00				; init rx buffer to $0100 - 255 byte buffer
	mov	RBIN, i
	mov	RBOUT, i
	mov	RBST, i				; buffer is empty to start
	ldi	i, 0x01
	mov	R9, i
	mov	R11, i					 

	andi	SPIST, 0xBF			; RX data rdy=0 

	ldi	i, (1<<RXEN)|(1<<TXEN)          ; |(1<<RXCIE)
	out	UCSRB, i			; Enable Receiver and Transmitter
	ret

;/******************************************************************************/
;/  PC Keyboard decoder															/
;/																				/
;/  Designed and written by Daryl Rictor (c)2004								/ 
;/******************************************************************************/
;
; All standard keys and control keys are decoded to 7 bit (bit 7=0) standard ASCII.
; Control key note: It is being assumed that if you hold down the ctrl key,
; you are going to press an alpha key (A-Z) with it (except break key defined below.)
; If you press another key, its ascii code's lower 5 bits will be send as a control
; code.  For example, Ctrl-1 sends $11, Ctrl-; sends $2B (Esc), Ctrl-F1 sends $01.
;
; The following non-standard keys are decoded with bit 7=1, bit 6=0 if not shifted,
; bit 6=1 if shifted, and bits 0-5 identify the key.
; 
; Function key translation:  
;              ASCII / Shifted ASCII
;            F1 - 81 / C1
;            F2 - 82 / C2
;            F3 - 83 / C3
;            F4 - 84 / C4
;            F5 - 85 / C5
;            F6 - 86 / C6
;            F7 - 87 / C7
;            F8 - 88 / C8
;            F9 - 89 / C9
;           F10 - 8A / CA
;           F11 - 8B / CB
;           F12 - 8C / CC
;
; The Print screen and Pause/Break keys are decoded as:
;                ASCII  Shifted ASCII
;    Ctrl-Break - 02       02  (Ctrl-B) (can be changed to AE/EE)(non-repeating key)  
;     Pause/Brk - 03       03  (Ctrl-C) (can change to 8E/CE)(non-repeating key)
;      Scrl Lck - 8D       CD  
;        PrtScn - 8F       CF
;
; The Alt key is decoded as a hold down (like shift and ctrl) but does not
; alter the ASCII code of the key(s) that follow.  Rather, it sends
; a Alt key-down code and a seperate Alt key-up code.  The user program
; will have to keep track of it if they want to use Alt keys. 
;
;      Alt down - A0
;        Alt up - E0
;
; Example byte stream of the Alt-F1 sequence:  A0 81 E0.  If Alt is held down longer
; than the repeat delay, a series of A0's will preceeed the 81 E0.
; i.e. A0 A0 A0 A0 A0 A0 81 E0.
;
; The special windows keys are decoded as follows:
;                           ASCII    Shifted ASCII
;        Left Menu Key -      A1          E1 
;       Right Menu Key -      A2          E2
;     Right option Key -      A3          E3
;            Power Key -      A4          E4  
;            Sleep Key -      A5          E5
;             Wake Key -      A6          E6
;
;
; The following "cursor" keys ignore the shift key and return their special key code 
; when numlock is off or their direct labeled key is pressed.  When numlock is on, the digits
; are returned reguardless of shift key state.        
; keypad(NumLck off) or Direct - ASCII    Keypad(NumLck on) ASCII
;          Keypad 0        Ins - 90                 30
;          Keypad .        Del - 7F                 2E
;          Keypad 7       Home - 97                 37
;          Keypad 1        End - 91                 31
;          Keypad 9       PgUp - 99                 39
;          Keypad 3       PgDn - 93                 33
;          Keypad 8    UpArrow - 98                 38
;          Keypad 2    DnArrow - 92                 32
;          Keypad 4    LfArrow - 94                 34
;          Keypad 6    RtArrow - 96                 36 
;          Keypad 5    (blank) - 95                 35
;

KB_INIT:						
	ldi	special, 1<<NUMLOCK		; init - num lock on, all other off

	ldi	i, 0x00
	out	GICR, i				; diable INT1
	ldi	i, 0x80
	out	GIFR, i				; clear active INT1 flags

; init kb buffer to $0200 - 255 byte buffer
	ldi	i, 0x00
	mov	KBIN, i
	mov	KBOUT, i
	mov	KBST, i
	ldi	i, 0x02
	mov	R13, i
	mov	R15, i 

	ldi	KBMODE, 0x00			; def mode is ASCII
	mov	KBEAT, KBMODE			; clr KBEAT
	andi	SPIST, 0x7F			; KB data rdy=0 

        ldi	i, 0xE0
	out	TCNT1H, i
	ldi	i, 0x00
	out	TCNT1L, i
  	ldi	i, 1 << TOIE1
	out	TIMSK, i
	SEI
	ldi	i, 0x05
	out	TCCR1B, i

init1:
	ldi	i, 0xff				; keybrd reset
	rcall	send       			; reset keyboard
	rcall	get				; 
	cpi	i, 0xFA				; ack?
	brne	init1				; resend reset cmd
 	rcall	get				; 
	cpi	i, 0xAA				; reset ok
	brne	init1				; resend reset cmd        
	rcall	SETLEDS				; set LED's and return

	CLI
  	ldi	i, 0x00
	out	TIMSK, i
	ldi	i, 0x00
	out	TCCR1B, i

	ldi	byte, 0x00			; clear scankey holder
	ldi	parity, 0x00			; clear parity holder
	ldi	pcnt, 0x00			; clear parity counter
	ldi	cntr, 0x0A			; bit counter 11 bits per frame  start, 8 data, parity, stop
	ldi	i, 0x80
	out	GIFR, i				; clear active INT1 flags
	out	GICR, i				; enable INT1
	cbi	ddrc, kbdat
	cbi	ddrd, kbclk			; set clk to input (change if port bits change)
	ret

TIM1_OVF:
	CLI
  	ldi	i, 0x00
	out	TIMSK, i
	ldi	i, 0x00
	out	TCCR1B, i
	ori	SPIST, 0x02			; set KB missing bit
	ldi	i, 0x04				; set stack pointer to top of SRAM 
	out	SPH, i				; 
	ldi	i, 0x5f				; 
	out	SPL, i				;
	rjmp	NOKB				; 

KBGET:
EXT_INT1:  ; EXT INT1 Service routine

	push	i
	in	i, sreg
	push	i

	cpi	cntr, 0x00			; stop bit
	breq	kbget6
	cpi	cntr, 0x01			; parity bit
	breq	kbget4

	cpi	cntr, 0x0A			; start bit
	breq	kbget8
		
	clc
	sbic	pinc, kbdat			; get data bit
	sec
	ror	byte  				; save bit to byte holder
	brpl	kbget3				; 
	inc	pcnt				; add 1 to parity counter
kbget3:
	dec	cntr				; dec bit counter
	pop	i
	out	sreg, i
	pop 	i
	reti					; 

kbget4: sbic	pinc, kbdat			; read kbdata for parity bit
	inc	parity				; if 1, set parity to 1        
	rjmp	kbget3				; finish

kbget6:
	sbis	pinc, kbdat			; get stop bit
	rjmp	kbgeterr			; error if stop is 0

	sbi	ddrd, kbclk			; clk = 0 : marked busy (change if port bits change)

	ldi	i, 0x00
	out	GICR, i				; disable INT1 until ready for next byte

	mov	i, pcnt				; get parity count
	andi	i, 0x01				; mask bit 1 only
	eor	i, parity			; compare with parity bit
	breq	kbgeterr			; bad parity
	rjmp	kbget3				; done, cntr = $FF data ready	

kbget8:
	sbis	pinc, kbdat			; read start bit
	rjmp	kbget3				; good start bit, finish

kbgeterr:

	ldi	i, 0x00
	out	GICR, i				; disable INT1 until ready for next byte

	ldi	i, 0xFE				; resend cmd
	rcall	send				; 

	ldi	byte, 0x00			; clear scankey holder
	ldi	parity, 0x00			; clear parity holder
	ldi	pcnt, 0x00			; clear parity counter
	ldi	cntr, 0x0A			; bit counter 11 bits per frame  start, 8 data, parity, stop

	ldi	i, 0x80
	out	GIFR, i				; clear active INT1 flags
	out	GICR, i				; enable INT1

	cbi	ddrc, kbdat
	cbi	ddrd, kbclk			; set clk to input (change if port bits change)
	pop	i
	out	sreg, i
	pop 	i
	reti					; 

KBDATRDY:
	sbrc    SPIST, 1			; if KB missing set, ignore KB
	ret					; missing, so return
	cpi	cntr, 0xFF
	breq	kbdatr1				; data, continue
	ret					; else return	
kbdatr1:
	tst	KBEAT				; need to eat codes?
	breq	kbdatra				; no
	dec	KBEAT				; yes, dec counter
	rjmp	kbdatr3				; flush code, get next

kbdatra: 
	mov	i, byte				; copy byte to i reg
	sbrs	KBMODE, 7			; KBMODE $00=ASCII, $80=SCANCODE
	rcall	convert				; if ASCII, call convert
	tst	i
	breq	kbdatr3				; if code is 0, ignore it
	inc	KBST
	brne	kbdatr2				; buffer not full 
	dec	KBST				; mark buffer full
	rjmp	kbdatr3				; get ready for next char
KBdatr2:
	ori	SPIST, 0x80			; KB data waiting 
	movw	YL, KBIN
	st	Y+, i
	mov	KBIN, YL

kbdatr3:
	ldi	byte, 0x00			; clear scankey holder
	ldi	parity, 0x00			; clear parity holder
	ldi	pcnt, 0x00			; clear parity counter
	ldi	cntr, 0x0A			; bit counter 11 bits per frame  start, 8 data, parity, stop
	ldi	i, 0x80
	out	GIFR, i				; clear active INT1 flags
	out	GICR, i				; enable INT1
	cbi	ddrc, kbdat
	cbi	ddrd, kbclk			; set clk to input (change if port bits change)
	ret					; exit

KB_RX_SPI:
	tst	KBST
	breq	KB_RX_SPI2
	movw	YL, KBOUT
	ld	i, Y+
	out	SPDR, i				; send to SPI 
	mov	KBOUT, YL
	dec	KBST
	brne	KB_RX_SPI1
	andi	SPIST, 0x7F			; KBRX data rdy=0 
KB_RX_SPI1:
	cbi	portc, IRQ
	ret
KB_RX_SPI2:
	ldi	i, 0x00
	out	SPDR, i
	cbi	portc, IRQ
	ret
;
; internal KB routines
;

SETLEDS:
	ldi	i, 0xED				; Set the keybrd LED's from kbleds variable
	rcall	send				; 
	rcall	get				;
 	cpi	i, 0xFA				; ack?
	brne   	setleds				; resend led cmd        
	mov   	i, special			; 
	andi    i, 0x07				; ensure bits 3-7 are 0
	rcall	send				; 
	rcall	get				; 
	cpi	i, 0xFA				; ack?
	brne	SETLEDS				; resend LED set cmd
	ret
;
HIGHLOW:
	sbis	pind, kbclk			
	rjmp	highlow				; wait while clk low
hl1: 
	sbic	pind, kbclk
	rjmp	hl1				; wait while clk is high
	clc					; clear Carry bit
	sbic	pinc, kbdat			; is data hi
	sec					; yes. set Carry bit
	ret					; exit with Carry = data input bit 
;
; Flush the keyboard buffer 
;
FLUSH:
	ldi	i, 0xf4				; flush buffer (fall through to Send command)
;
; send a byte to the keyboard from i
;
SEND:
	mov	byte, i				; byte to send into Byte
	mov	lastbyte, i			; copy to lastbyte also in case of send error
	sbi	ddrd, kbclk			; clk & dat as outputs
	ldi	i, 0xFA				; 250 dec * 4 clks = 100uS
sendw:						;
	dec	i		            	; 
	nop                         		;
	brne	sendw				; 100uS delay
	ldi	parity, 0x00			; parity counter
	ldi	cntr, 0x08			; bit counter 
	sbi	ddrc, kbdat			; set data low
	cbi	ddrd, kbclk			; set clk input
	rcall	highlow				; wait for clock transition from kb and read data
send1:
	lsr   	byte				; get lsb first
	brcs  	mark				; 
	sbi	ddrc, kbdat			; set data line low
	rjmp	next				; 
mark:
	cbi	ddrc, kbdat			; set data line high
	inc     parity				; inc parity counter
next:
	rcall	highlow				; 
	dec  	cntr  				; bit counter
	brne    send1				; send 8 data bits

	mov	i, parity
	andi	i, 0x01				; get odd or even
	brne	pclr				; if odd, send 0
	cbi	ddrc, kbdat			; set data bit high
	rjmp	ack				; ack
pclr:
	sbi	ddrc, kbdat			; set data bit low
ack:
	rcall	highlow         		; 
	cbi	ddrd, kbclk			; set clk & data to input
	cbi	ddrc, kbdat			;	
	rcall	highlow				; wait for ack from keyboard
	brcc    send2
	rjmp   	KB_Init				; VERY CRUDE error handler - re-init the keyboard
send2:
	sbis	pind, kbclk			; wait for clk to go high
	rjmp	send2				;
	rjmp	disable				; done
;
;
; Error sends a resend command and then gets the code again
;
error:
	ldi	i, 0xFE				; resend cmd
	rcall	send				; 
;
; GET waits for one scancode from the keyboard, save in byte
;
GET:
	ldi	byte, 0x00			; clear scankey holder
	ldi	parity, 0x00			; clear parity holder
	ldi	pcnt, 0x00			; clear parity counter
	ldi	cntr, 0x08			; bit counter 
	cbi	ddrc, kbdat
	cbi	ddrd, kbclk			; set clk to input (change if port bits change)
	nop					;
	nop					;
	nop					; delay needed to give clk time to rise
get1:
	sbic	pind, kbclk
	rjmp	get1				; wait while clk is high

	sbic	pinc, kbdat			; get start bit, if 1, false start bit, do again 
	rjmp	get1
get2:
	rcall	highlow				; wait for clk to return high then go low again
	ror	byte  				; save bit to byte holder
	brpl	get3				; 
	inc	pcnt				; add 1 to parity counter
get3:
	dec	cntr				; dec bit counter
	brne	get2   				; get next bit if bit count > 0 
	rcall	highlow				; wait for parity bit
	brcc	get4				; if parity bit 0 do nothing
	inc	parity				; if 1, set parity to 1        
get4:
	mov	i, pcnt				; get parity count
	eor	i, parity			; compare with parity bit
	andi	i, 0x01				; mask bit 1 only
	breq	error				; bad parity
	rcall	highlow				; wait for stop bit
	brcc	error				; 0=bad stop bit 
	cpi	byte, 0x00			; is byte rcv'd =0?
	breq	get				; no data, do again
	rcall	disable				; 
	mov	i, byte				; 
	ret					; 
;
; disable keyboard from sending
;
disable:					; disable kb from sending more data
	sbi	ddrd, kbclk			; clk = 0 (change if port bits change)
	ret					; 
;
; convert scancode to an ASCII code
;
CONVERT:
	sbrc	special, rlse			; was a release code pending?
	rjmp	Rlsproc				; yes - process this key
	sbrc	special, ext			; was extended code entered?
	rjmp	Extproc				; yes, do further tests
	cpi	byte, 0x12			; no, is it the L shift key? 
	breq	shftky
	cpi	byte, 0x59			; is it the R shift key? 
	breq	shftky				; 
exttst:						;***** do not add test above this line without changing Extproc code!
	cpi	byte, 0x14			; is it the ctrl key? 
	breq	ctrlky				; 
	cpi	byte, 0x58			; is it the capslock? 
	breq	caplky				; 
	cpi	byte, 0x77			; is it the numlock key? 
	breq	numlky				; no, test code under 128
	cpi	byte, 0x7E			; scroll Lock key?
	breq    scrlky				; yes
	cpi	byte, 0x7F			; is it < 0x80
	brcs    lookup				; yes, do direct conversion to ASCII
	cpi	byte, 0x83			; is it the F7 key?
	breq	F7ky				; 
	cpi	byte, 0xE0			; is it the E0 code
	breq	extky				; 
	cpi	byte, 0xE1			; is it pause/break?
	breq	brkky				; 
	cpi	byte, 0xF0			; is it the key release code?
	brne	exttst1                         ; no, ignore and get next key
	sbr	special, 1<<rlse		;
exttst1:
	clr	i
	ret

lookup:
	ldi	ZH, 0x1E			; set page base for index table (pre *2)
	mov	ZL, byte			; set code to index
	lsl	ZL				; *2 for lookup in word table unshifted byte
	sbrc	special, shift
	sbr	ZL, 1				; set shifted byte in word table	
	sbrc	special, extsh
	sbr	ZL, 1				; set shifted byte in word table	
	cpi	byte, 0x78			; is it F11?
	breq	lookup1				; yes, just get it
	cpi	byte, 0x7E			; is it scroll lock?
	breq	lookup1				; yes, just get it
	cpi	byte, 0x69			; test for numpad?
	brcs	lookup1				; not numpad
	sbrs	special, numlock		; test for numlock off
	sbr	ZL, 1				; if off, get shifted code from table
lookup1:
	lpm	i, Z				; lookup the ASCII code from the scan code
	sbrc	special, control		; is control key active?
	andi	i, 0x1F				; mask out control key codes from 0x00-0x1F
	sbrc	special, rlse			; is this a release code?
	andi	i, 0x00				; clear it out
	sbrs	special, capslock		; is capslock set
	rjmp	final				; no
	cpi	i, 0x61				; <"A"
	brcs	final
	cpi	i, 0x7B				; >"Z"+1
	brcc	final
	andi	i, 0xDF				; set letter to CAPS
final:
	ret					; done. exit with ASCII in i
	
shftky:
	sbr	special, 1<<shift		; set shift bit
	clr	i
	ret
ctrlky:
	sbr	special, 1<<control		; set control bit
	clr	i
	ret
caplky:
	ldi	i, 1<<capslock
	eor	special, i			; toggle the capslock
	rcall	setleds				; update the LED's
	clr	i
	ret
numlky:	
	ldi	i, 1<<numlock
	eor	special, i			; toggle the numlock
	rcall	setleds				; update the LED's
	clr	i
	ret
scrlky:	
	ldi	i, 1<<scrllock
	eor	special, i			; toggle the scroll-lock
	rcall	setleds				; update the LED's
        ldi	i, 0x7E                         ;
        mov     byte, i                         ; 
	rjmp	lookup				; send scroll lock code

F7ky:	
	ldi	byte, 0x02			; yes F7, set scancode to modified F7 code 0x02
	rjmp	lookup				; process it
extky:
	sbr	special, 1<<ext			; set the extend flag in the flag reg
	clr	i
	ret
brkky:
	ldi	i, 0x07
	mov	KBEAT, i
	ldi	i, 0x03				; set Pause/break code
	ret					; return it
ctrlbrk:
	ldi	i, 0x02				; set Control- Pause/break code
	ret					; return it
Extshky:
	sbr	special, 1<<extsh		; set ext shift bit
	clr	i
	ret
Prtscnky:
	cbr	special, 1<<extsh		; turn off ext shift flag
	ldi	byte, 0x0F			; yes, set new scan code for PrtScrn
	rjmp	lookup				; and process it
KPdivky:
	ldi	i, 0x2F				; yes, set ASCII code for /
	ret					; and send it to host
Rlsalt:
	ldi	i, 0xE0				; alt Rls ASCII Code
	ret
Rlsshft:
	cbr	special, 1<<shift		; drop shift mode
rlsextsh:
	cbr	special, 1<<rlse		; yes - reset flag
	cbr	special, 1<<extsh		; clear ext shift register bit
	clr	i
	ret
rlsctrl:
	cbr	special, 1<<control		; drop control mode
	clr	i
	ret

Extproc:
	cbr	special, 1<<ext			; turn off ext flag
	cpi	byte, 0x12			; was the E012 code entered?
	breq	Extshky
	cpi	byte, 0x7C			; was it E07C - Print Screen?
	breq	Prtscnky
	cpi	byte, 0x4A			; was it KP /?
	breq	KPdivky	
	cpi	byte, 0xF0			; is it a rlse code?
	breq	Extrls				; 
	cpi	byte, 0x7E			; is it the ctrl-brk key?
	breq	Ctrlbrk				;
	cpi	byte, 0x68			; test for cursor/movement keys (E069 - E07D)
	brcc	extp1				; do lookup
	rjmp	Exttst				; 
Extp1:
	rjmp	lookup

Extrls: 
	sbr	special, 1<<rlse
	clr	i
	ret

Rlsproc:
	sbrs	special, ext			; was extended code entered?
	rjmp	Rlsproc1			;
	cpi	byte, 0x12			; was it ext shift?
	breq	rlsextsh			; 
Rlsproc1:
	cbr	special, 1<<rlse		; yes - reset flag
	cpi	byte, 0x11			; was alt released?
	breq	Rlsalt
	cpi	byte, 0x12			; left shift
	breq	Rlsshft
	cpi	byte, 0x59			; right shift
	breq	Rlsshft
	cpi	byte, 0x14			; control key
	breq	rlsctrl				; 
	clr	i
	ret					; don't care about the rest, get next key
;
.cseg
.org	0xF00
scancode:
;               norm    shift     scancode
	 .db		$00,	$00		; 00 no key pressed	
	 .db		$89,	$C9		; 01 F9	
	 .db		$87,	$C7		; 02 relocated F7	
	 .db		$85,	$C5		; 03 F5	
	 .db		$83,	$C3		; 04 F3	
	 .db		$81,	$C1		; 05 F1	
	 .db		$82,	$C2		; 06 F2	
	 .db		$8C,	$CC		; 07 F12	
	 .db		$00,	$00		; 08	
	 .db		$8A,	$CA		; 09 F10	
	 .db		$88,	$C8		; 0A F8	
	 .db		$86,	$C6		; 0B F6	
	 .db		$84,	$C4		; 0C F4	
	 .db		$09,	$09		; 0D tab	
	 .db		$60,	$7E		; 0E `~	
	 .db		$8F,	$CF		; 0F relocated Print Screen key	
	 .db		$03,	$03		; 10 relocated Pause/Break key	
	 .db		$A0,	$A0		; 11 left alt (right alt too)	
	 .db		$00,	$00		; 12 left shift	
	 .db		$00,	$00		; 13 
	 .db		$00,	$00		; 14 left ctrl (right ctrl too)	
	 .db		$71,	$51		; 15 qQ	
	 .db		$31,	$21		; 16 1!	
	 .db		$00,	$00		; 17	
	 .db		$00,	$00		; 18	
	 .db		$00,	$00		; 19	
	 .db		$7A,	$5A		; 1A zZ	
	 .db		$73,	$53		; 1B sS	
	 .db		$61,	$41		; 1C aA	
	 .db		$77,	$57		; 1D wW	
	 .db		$32,	$40		; 1E 2@	
	 .db		$A1,	$E1		; 1F Windows 98 menu key (left side)	
	 .db		$00,	$00		; 20 	
	 .db		$63,	$43		; 21 cC	
	 .db		$78,	$58		; 22 xX	
	 .db		$64,	$44		; 23 dD	
	 .db		$65,	$45		; 24 eE	
	 .db		$34,	$24		; 25 4$	
	 .db		$33,	$23		; 26 3#	
	 .db		$A2,	$E2		; 27 Windows 98 menu key (right side)	
	 .db		$00,	$00		; 28	
	 .db		$20,	$20		; 29 space	
	 .db		$76,	$56		; 2A vV	
	 .db		$66,	$46		; 2B fF	
	 .db		$74,	$54		; 2C tT	
	 .db		$72,	$52		; 2D rR	
	 .db		$35,	$25		; 2E 5%	
	 .db		$A3,	$E3		; 2F Windows 98 option key (right click, right side)	
	 .db		$00,	$00		; 30	
	 .db		$6E,	$4E		; 31 nN	
	 .db		$62,	$42		; 32 bB	
	 .db		$68,	$48		; 33 hH	
	 .db		$67,	$47		; 34 gG	
	 .db		$79,	$59		; 35 yY	
	 .db		$36,	$5E		; 36 6^	
	 .db		$A4,	$E4		; 37 Power Key 	
	 .db		$00,	$00		; 38	
	 .db		$00,	$00		; 39	
	 .db		$6D,	$4D		; 3A mM	
	 .db		$6A,	$4A		; 3B jJ	
	 .db		$75,	$55		; 3C uU	
	 .db		$37,	$26		; 3D 7&	
	 .db		$38,	$2A		; 3E 8*	
	 .db		$A5,	$E5		; 3F Sleep Key	
	 .db		$00,	$00		; 40	
	 .db		$2C,	$3C		; 41 ,<	
	 .db		$6B,	$4B		; 42 kK	
	 .db		$69,	$49		; 43 iI	
	 .db		$6F,	$4F		; 44 oO	
	 .db		$30,	$29		; 45 0)	
	 .db		$39,	$28		; 46 9(	
	 .db		$00,	$00		; 47	
	 .db		$00,	$00		; 48	
	 .db		$2E,	$3E		; 49 .>	
	 .db		$2F,	$3F		; 4A /?	
	 .db		$6C,	$4C		; 4B lL	
	 .db		$3B,	$3A		; 4C ;:	
	 .db		$70,	$50		; 4D pP	
	 .db		$2D,	$5F		; 4E -_	
	 .db		$00,	$00		; 4F	
	 .db		$00,	$00		; 50	
	 .db		$00,	$00		; 51	
	 .db		$27,	$22		; 52 '"	
	 .db		$00,	$00		; 53	
	 .db		$5B,	$7B		; 54 [{	
	 .db		$3D,	$2B		; 55 =+	
	 .db		$00,	$00		; 56	
	 .db		$00,	$00		; 57	
	 .db		$00,	$00		; 58 caps	
	 .db		$00,	$00		; 59 r shift	
	 .db		$0D,	$0D		; 5A <Enter>	
	 .db		$5D,	$7D		; 5B ]}	
	 .db		$00,	$00		; 5C	
	 .db		$5C,	$7C		; 5D \|	
	 .db		$A6,	$E6		; 5E Wake Key	
	 .db		$00,	$00		; 5F	
	 .db		$00,	$00		; 60	
	 .db		$00,	$00		; 61	
	 .db		$00,	$00		; 62	
	 .db		$00,	$00		; 63	
	 .db		$00,	$00		; 64	
	 .db		$00,	$00		; 65	
	 .db		$08,	$08		; 66 bkspace	
	 .db		$00,	$00		; 67	
	 .db		$00,	$00		; 68	
	 .db		$31,	$91		; 69 kp 1 {End}
	 .db		$2f,	$2f		; 6A kp / converted from E04A in code	
	 .db		$34,	$94		; 6B kp 4 {Left}	
	 .db		$37,	$97		; 6C kp 7 {Home}	
	 .db		$00,	$00		; 6D	
	 .db		$00,	$00		; 6E	
	 .db		$00,	$00		; 6F	
	 .db		$30,	$90		; 70 kp 0 {Ins}	
	 .db		$2E,	$7F		; 71 kp . {Del}	
	 .db		$32,	$92		; 72 kp 2 {Down}	
	 .db		$35,	$95		; 73 kp 5	
	 .db		$36,	$96		; 74 kp 6 {Right}	
	 .db		$38,	$98		; 75 kp 8 {Up}	
	 .db		$1B,	$1B		; 76 esc	
	 .db		$00,	$00		; 77 num lock	
	 .db		$8B,	$CB		; 78 F11	
	 .db		$2B,	$2B		; 79 kp +	
	 .db		$33,	$93		; 7A kp 3 {PgDn}	
	 .db		$2D,	$2D		; 7B kp -	
	 .db		$2A,	$2A		; 7C kp *	
	 .db		$39,	$99		; 7D kp 9 {PgUp}	
	 .db		$8D,	$CD		; 7E scroll lock	
	 .db		$00,	$00		; 7F

.cseg
.org	0xFFF
CALBYTE:
	.db		$9F,	$80		; 8MHZ calibration byte

;
; End of Program
;
