; /////////////////////////////////////////////////////////////////////////////////
;//                                                                               //
;//  This program and its associated documentation are provided for your personal //
;//  use only and appear here exclusively by permission of the copyright holder.  //
;//  Please contact the copyright holder before re-distributing, re-publishing    //
;//  or disseminating this copyrighted work. This code is not GPL or in the       //
;//  public domain. Please respect the author"s copyright.                        //
;//                                                                               //
;//  No waranty, either expressed or implied, are given.  I assume no liability   //
;//  for its use in any project or device.                                        //
;//                                                                               //
;//  Your use of this program indicates your acceptance of all license terms.     //
;//  This particular version is freeware as long as the copyright messages are    //
;//  left intact.                                                                 //
;//                                                                               //
; /////////////////////////////////////////////////////////////////////////////////
;
;*******************************************************************************
;
;  65SPI and ATMega8 Interface Routines for SBC-3  (c) Daryl Rictor Nov 6, 2008
;
;*******************************************************************************
;

;*********************** I/O Locations *******************************
;
; VIA1 connection definitions for ATMega8
VIA1_PCR	=	$0C 
VIA1_IFR	=	$0D
VIA1_IER	=	$0E
VIA1_ORB	=	$00

; 65SPI register addresses (using a modified 65SPI for SBC-3)
SPIDR		=	$20			; SPI data port
SPISR		=	$21			; SPI status
SPICR		=	$21			; SPI control
SPIDIV		=	$22			; SPI SCLK divisor (6 bits)
SPISSR		=	$23			; SPI Slave Select register
SPIISR		=	$24			; SPI Interrupt status register

;*********************** Register Definitions *******************************

; SPI command structure
; $00 - null (used to receive data)
; $01 - Get Status - see status register bit definitions
; $02 - Get RS-232 RX byte
; $03 - send RS-232 TX byte
; $04 - Get RS-232 parameters
; $05 - Set RS-232 parms : bit 0-2 = speed, bit 3 = 7/8 bit, bit 4 = parity y/n, bit 5 = parity odd/even, bit 6 = stop 1/2
; $06 - get KB byte 
; $07 - send kb cmd	0= ASCII mode, 1= scancode mode, 2= future > 2 = KB CMD byte
; $08 - init RS232 - Set to default settings
; $09 - init KB - Re-init kb and set to ASCII mode
;
; 65SPI Status Register bit definitions  
; 7= IRQ - 1 = operation complete, data waiting  0 = no attention needed
; 6= IER - 1 = Enable interrupt output, 0= disable interrupt output
; 5= BSY - 1 = shifting in progress, 0= shifting idle
; 4= FRX - 1 = Fast receive, reading SPIDR initiates a new shift; 0 = fast receive disabled
; 3= TMO - 1 = put MOSI pin in hi-z, 0 = MOSI output enabled 
; 2= ECE - 1 = External clock enabled (SS7 is an input), 0 = ext clock disabled SS7 is an output
; 1= CPOL - Clock Polarity } selects shift mode 0-3			 
; 0= CPHA - Clock phase	 }
;
; 65SPI Control Register bit definitions
; 7= not used
; 6= IER - 1 = Enable interrupt output, 0= disable interrupt output
; 5= not used
; 4= FRX - 1 = Fast receive, reading SPIDR initiates a new shift; 0 = fast receive disabled
; 3= TMO - 1 = put MOSI pin in hi-z, 0 = MOSI output enabled 
; 2= ECE - 1 = External clock enabled (SS7 is an input), 0 = ext clock disabled SS7 is an output
; 1= CPOL - Clock Polarity } selects shift mode 0-3			 
; 0= CPHA - Clock phase	 }
;
;ATM8 Status bit definitions
; 7= KB data waiting
; 6= RX data waiting
; 5= TX Ready
; 4= RX Frame error
; 3= RX overrun
; 2= RX parity
; 1= KB Missing
; 0= not used
;
;*********************** SPI Interface Routines *******************************
SPI_init	LDA	Via1_pcr		;
		and	#$0F			;
		STA	Via1_pcr		; set CB1 to neg flag
SPI_init1	lda	Via1_IFR		;
		and	#$10			; wait for CB1 to fall
		beq	SPI_init1		; wait for ATM8 to signal ready
		lda	#$FF			;
		sta	SPISSR			; init Slave Select register, all low except SS6 high
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	#$00			;
		sta	SPICR			; control reg - no IRQ, int clock, mode 0
.if PAL=1
		lda	#$02			; PAL SCLK divisor
.else
		lda	#$01			; NTSC SCLK divisor
.fi
		sta	SPIDIV			; sclk divisor (ATM8 max is 2MHz) NTSC=1, PAL=2
		lda	SPIDR			; clear data register flag
		rts				

GET_STATUS	lda	#$40			;
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag		
		jsr	SPI_STATUS		; get status
		pha				; save status
		lda	#$40			; 
		TSB	SPISSR			; deselect ATM8
		pla				; restore status
		rts				;

SPI_STATUS	lda	#$01			; GET ATM8 STATUS
		sta	SPIDR			; send get Status cmd
SPI_STAT1	lda	Via1_IFR		;
		and	#$10			; CB1 mask
		beq	SPI_STAT1		; wait for ATM8 to signal ready
		lda	SPIDR			; clear flag
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	#$00			; null cmd
		sta	SPIDR			; send data to kb
SPI_STAT2	lda	SPISR				
		bpl	SPI_STAT2		; wait for tx to end
		lda	SPIDR			; read SPI status
		rts

;----------------------------------------------------------------------------
SPIKB_init	lda	#$40
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag
		jsr	SPI_Status		;
		and	#$02			; mask KB missing
		beq	SPIKB_init9		; no errors found
		brk				; stop - KB missing error
		.byte 	$EE			; KB missing
SPIKB_init9	lda	#$40			; *** ADD code here to set ASCII mode
		tsb	SPISSR			; deselect ATM8
		rts				; done

SET_KB			pha			; save  A
SET_KB0		lda	#$40
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag
		lda	#$07			;
		sta	SPIDR			; send Send KB data cmd
SET_KB1		lda	Via1_IFR		;
		and	#$10			; CB1 mask
		beq	SET_KB1			; wait for tx to end
		lda	SPIDR			; clear data ready flag		
		lda	Via1_ORB		; clear Via1 CB1 flag
		pla				; 
		sta	SPIDR			; send data to kb
SET_KB2		lda	SPISR			;
		bpl	SET_KB2			; wait for tx to end
		lda	SPIDR			; read SPI data reg to clear flags
		lda	#$40			; 
		tsb	SPISSR			; deselect ATM8
		rts

INPUT_CHR
SPI_KBIN	lda	#$40			;
		TRB	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear Via1 CB1 flag
SPI_KBIN1	jsr	SPI_STATUS		; get status
		bpl	SPI_KBIN1		; keep testing for KB data
SPI_KBIN2	lda	#$06
		sta	SPIDR			; send get KB data
SPI_KBIN3	lda	Via1_IFR		;
		and	#$10
		beq	SPI_KBIN3		; wait for ATM8 to signal ready
		lda	SPIDR			; clear data ready flag		
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	#$00			; null cmd
		sta	SPIDR			; send null to get data
SPI_KBIN4	lda	SPISR			; 
		bpl	SPI_KBIN4		; wait for tx to end
		lda	SPIDR			; read KB data
		pha
		lda	#$40			; 
		TSB	SPISSR			; deselect ATM8
		pla
		sec
		rts
					
SCAN_INPUT
SPI_KBSCAN	lda	#$40			;
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear Via1 CB1 flag
		jsr	SPI_Status		; get status
		bmi	SPI_KBIN2		; if byte ready, get it and return w/carry set				
		lda	#$40			; 
		tsb	SPISSR			; deselect ATM8
		clc				; otherwise, return with carry clear
		rts					
;----------------------------------------------------------------------------

SPISER_init	lda	#$40
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag
		jsr	SPI_Status		;
		and	#$7C			; mask SER flags
		cmp	#$20			; TX ready flag set
		beq	SPISER_init9		; no errors found
		brk				; stop
		.byte	$CC			; RS232 error
SPISER_init9	lda	#$40			;*** ADD code to set comm mode?		 
		tsb	SPISSR			; deselect ATM8
		rts				;

SCAN_SERIN	lda	#$40			;
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear Via1 CB1 flag
		jsr	SPI_Status		; get status
		and	#$40			; RX data ready
		sec
		bne	SCAN_SERIN1		; if byte ready, return w/carry set				
		clc				; otherwise, return with carry clear
SCAN_SERIN1	lda	#$40			; 
		tsb	SPISSR			; deselect ATM8
		rts				  

SER_IN		lda	#$40			;
		TRB	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear Via1 CB1 flag
SER_IN1		jsr	SPI_STATUS		; get status
		and	#$40			; RX data ready
		beq	SER_IN1			; keep testing for RS-232 data
SER_IN2		lda	#$02
		sta	SPIDR			; send get Rs-232 RX data
SER_IN3		lda	Via1_IFR
		and	#$10
		beq	SER_IN3			; wait for ATM8 to signal ready
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	SPIDR			; clear data ready flag		
		lda	#$00			; null cmd
		sta	SPIDR			; send null to get data
SER_IN4		lda	SPISR			; 
		bpl	SER_IN4			; wait for tx to end
		lda	SPIDR			; read KB data
		pha
		lda	#$40			; 
		TSB	SPISSR			; deselect ATM8
		pla
		sec
		rts

SER_OUT		pha				; called with data in A
		lda	#$40			;
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag
SER_OUT0	jsr	SPI_Status		; get status
		and	#$20			;
		beq	SER_OUT0		; wait for tx to be empty
		lda	#$03			;
		sta	SPIDR			; send RS-232 data
SER_OUT1	lda	Via1_IFR		;
		and	#$10			; CB1 mask
		beq	SER_OUT1		; wait for tx to end
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	SPIDR			; clear data ready flag		
		pla				; 
		sta	SPIDR			; send data to SER port
SER_OUT2	lda	SPISR			;
		bpl	SER_OUT2		; wait for tx to end
		lda	SPIDR			; read SPI data reg to clear flags
		lda	#$40			; 
		tsb	SPISSR			; deselect ATM8
		rts

GET_SER		lda	#$40			; Get Serial port Parameters
		TRB	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear Via1 CB1 flag
GET_SER2	lda	#$04
		sta	SPIDR			; send get KB data
GET_SER3	lda	Via1_IFR		;
		and	#$10			;
		beq	GET_SER3		; wait for ATM8 to signal ready
		lda	SPIDR			; clear data ready flag		
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	#$00			; null cmd
		sta	SPIDR			; send null to get data
GET_SER4	lda	SPISR			; 
		bpl	GET_SER4		; wait for tx to end
		lda	SPIDR			; read KB data
		pha				; save parm in stack
		lda	#$40			; 
		TSB	SPISSR			; deselect ATM8
		pla				; return parm from stack
		rts				; return with parm in A

SET_SER		pha				; called with config in A
		lda	#$40			;
		trb	SPISSR			; select ATM8 SS
		lda	Via1_ORB		; clear VIA1 flag
		lda	#$05			;
		sta	SPIDR			; send RS-232 config cmd
SET_SER1	lda	Via1_IFR		;
		and	#$10			; CB1 mask
		beq	SET_SER1		; wait for tx to end
		lda	Via1_ORB		; clear Via1 CB1 flag
		lda	SPIDR			; clear data ready flag		
		pla				; 
		sta	SPIDR			; send data to SER port
SET_SER2	lda	SPISR			;
		bpl	SET_SER2		; wait for tx to end
		lda	SPIDR			; read SPI data reg to clear flags
		lda	#$40			; 
		tsb	SPISSR			; deselect ATM8
		rts
