; /////////////////////////////////////////////////////////////////////////////////
;//                                                                               //
;//  This program and its associated documentation are provided for your personal //
;//  use only and appear here exclusively by permission of the copyright holder.  //
;//  Please contact the copyright holder before re-distributing, re-publishing    //
;//  or disseminating this copyrighted work. This code is not GPL or in the       // 
;//  public domain. Please respect the author's copyright.                        //
;//                                                                               //
;//  No waranty, either expressed or implied, are given.  I assume no liability   //
;//  for its use in any project or device.                                        //
;//                                                                               //
;//  Your use of this program indicates your acceptance of all license terms.     //
;//  This particular version is freeware as long as the copyright messages are    //
;//  left intact.                                                                 //  
;//                                                                               //
; /////////////////////////////////////////////////////////////////////////////////
;
; XMODEM/CRC Sender/Receiver for the 65C816 running in Native Mode
;
; (c) Daryl Rictor Oct 6, 2008
;
; Version 1.0
;
; A simple file transfer program to allow transfers between the SBC and a 
; console device utilizing the x-modem/CRC transfer protocol.  Requires just
; under 1.5k of either RAM or ROM, 132 bytes of RAM for the receive buffer,
; and 12 bytes of zero page RAM for variable storage.
;
;**************************************************************************
; This implementation of XMODEM/CRC does NOT conform strictly to the 
; XMODEM protocol standard in that it (1) does not accurately time character
; reception or (2) fall back to the Checksum mode.
;
; (1) For timing, it uses a crude timing loop to provide approximate
; delays.  These have been calibrated against a 1MHz CPU clock.  I have
; found that CPU clock speed of up to 5MHz also work but may not in
; every case.  Windows HyperTerminal worked quite well at both speeds!
;
; (2) Most modern terminal programs support XMODEM/CRC which can detect a
; wider range of transmission errors so the fallback to the simple checksum
; calculation was not implemented to save space.
;
;**************************************************************************
; Files transferred via XMODEM-CRC will have the load address contained in
; the first two bytes in little-endian format:  
;  FIRST BLOCK
;     offset(0) = lo(load start address),
;     offset(1) = hi(load start address)
;     offset(2) = data byte (0)
;     offset(n) = data byte (n-2)
;
; Subsequent blocks
;     offset(n) = data byte (n)
;
; One note, XMODEM send 128 byte blocks.  If the block of memory that
; you wish to save is smaller than the 128 byte block boundary, then
; the last block will be padded with zeros.  Upon reloading, the
; data will be written back to the original location.  In addition, the
; padded zeros WILL also be written into RAM, which could overwrite other
; data.   
;
;-------------------------- The Code ----------------------------
;
; zero page variables (adjust these to suit your needs)
;
SourcePg       =     $E0               ; start block of send
TargetPg       =     $E1               ; end block of send, destination block of receive
CurrPg         =     $E4               ; current page of send data
lastblk        =     $E5               ; flag for last block
blkno          =     $E6               ; block number 
errcnt         =     $E3               ; error counter 10 is the limit
bflag          =     $E2               ; block flag 
crc            =     $E7               ; CRC lo byte  (two byte variable)
crch           =     $E8               ; CRC hi byte  
ptr            =     $E9               ; data pointer (two byte variable)
ptrh           =     $EA               ; = Addrptr in Monitor
eofp           =     $EB               ; end of file address pointer (2 bytes)
eofph          =     $EC               ; = Hexdigits in Monitor 
retry          =     $EE               ; retry counter 
retry2         =     $EF               ; 2nd counter
;
; non-zero page variables and buffers
;
Rbuff          =     $0200             ; temp 132 byte receive buffer 
                                       ;(place anywhere, page aligned)
;
;
; XMODEM Control Character Constants
SOH            =     $01               ; start block
EOT            =     $04               ; end of text marker
ACK            =     $06               ; good block acknowledged
NAK            =     $15               ; bad block acknowledged
CAN            =     $18               ; cancel (not standard, not supported)
CR             =     $0d               ; carriage return
LF             =     $0a               ; line feed
ESC            =     $1b               ; ESC to exit

;
;^^^^^^^^^^^^^^^^^^^^^^ Start of Program ^^^^^^^^^^^^^^^^^^^^^^
;
; Xmodem/CRC transfer routines
; By Daryl Rictor, August 8, 2002
;
; v1.0  released on Aug 8, 2002.
; v1.1  updated for 65816 Native Mode - Nov 5, 2008
;
;*************************************************************************
; Enter this routine with the beginning address stored in the zero page address
; pointed to by ptr & ptrh and the ending address stored in the zero page address
; pointed to by eofp & eofph.
;
;
XModemSend     jsr   PrintMsg          ; send prompt and info
               lda   #$00              ;
               sta   errcnt            ; error counter set to 0
               sta   lastblk           ; set flag to false
               lda   #$01              ;
               sta   blkno             ; set block # to 1
Wait4CRC       lda   #$ff              ; 3 seconds
               sta   retry2            ;
               jsr   GetByte           ;
               bcc   Wait4CRC          ; wait for something to come in...
               cmp   #"C"              ; is it the "C" to start a CRC xfer?
               beq   SetstAddr         ; yes
               cmp   #ESC              ; is it a cancel? <Esc> Key
               bne   Wait4CRC          ; No, wait for another character
               jmp   PrtAbort          ; Print abort msg and exit
SetstAddr      ldy   #$00              ; init data block offset to 0
               ldx   #$04              ; preload X to Receive buffer
               lda   #$01              ; manually load blk number        
               sta   Rbuff             ; into 1st byte
               lda   #$FE              ; load 1's comp of block #        
               sta   Rbuff+1           ; into 2nd byte
               lda   ptr               ; load low byte of start address                
               sta   Rbuff+2           ; into 3rd byte        
               lda   ptrh              ; load hi byte of start address                
               sta   Rbuff+3           ; into 4th byte
               lda   SourcePg          ; get start block
               sta   CurrPg            ; set current page
               bra   Ldbuff1           ; jump into buffer load routine

LdBuffer       lda   Lastblk           ; Was the last block sent?
               beq   LdBuff0           ; no, send the next one        
               jmp   Done              ; yes, we're done
LdBuff0        ldx   #$02              ; init pointers
               ldy   #$00              ;
               inc   Blkno             ; inc block counter
               lda   Blkno             ; 
               sta   Rbuff             ; save in 1st byte of buffer
               eor   #$FF              ; 
               sta   Rbuff+1           ; save 1's comp of blkno next
LdBuff1        lda   CurrPg            ; get current page
               pha                     ; put on stack
               #PLB                    ; set data page      
               lda   (ptr),y           ; 
               #STA_ALX  Rbuff,$00     ; save 128 bytes of data
               lda   #$00              ; 
               pha                     ; put on stack
               #PLB                    ; set data page back to 0
LdBuff2        lda   TargetPg          ; get last block
               cmp   CurrPg            ; equal current block?
               bne   LdBuff4           ; no, inc pointer and continue
               sec                     ; 
               lda   eofp              ;
               sbc   ptr               ; Are we at the last address?
               bne   LdBuff4           ; no, inc pointer and continue
               lda   eofph             ;
               sbc   ptrh              ;
               bne   LdBuff4           ; 
               inc   LastBlk           ; Yes, Set last byte flag
LdBuff3        inx                     ;
               cpx   #$82              ; Are we at the end of the 128 byte block?
               beq   SCalcCRC          ; Yes, calc CRC
               lda   #$00              ; Fill rest of 128 bytes with $00
               sta   Rbuff,x           ;
               beq   LdBuff3           ; Branch always
LdBuff4        inc   ptr               ; Inc address pointer
               bne   LdBuff5           ;
               inc   ptrh              ;
               bne   LdBuff5           ;
               inc   CurrPg            ; go to next block
LdBuff5        inx                     ;
               cpx   #$82              ; last byte in block?
               bne   LdBuff1           ; no, get the next
SCalcCRC       jsr   CalcCRC
               lda   crch              ; save Hi byte of CRC to buffer
               sta   Rbuff,y           ;
               iny                     ;
               lda   crc               ; save lo byte of CRC to buffer
               sta   Rbuff,y           ;
Resend         ldx   #$00              ;
               lda   #SOH              ;
               jsr   Put_chr           ; send SOH
SendBlk        lda   Rbuff,x           ; Send 132 bytes in buffer to the console
               jsr   Put_chr           ;
               inx                     ;
               cpx   #$84              ; last byte?
               bne   SendBlk           ; no, get next
               lda   #$FF              ; yes, set 3 second delay 
               sta   retry2            ; and
               jsr   GetByte           ; Wait for Ack/Nack
               bcc   Seterror          ; No chr received after 3 seconds, resend
               cmp   #ACK              ; Chr received... is it:
               bne   TestNAK           ;
               jmp   LdBuffer          ; ACK, send next block
TestNAK        cmp   #NAK              ; 
               beq   Seterror          ; NAK, inc errors and resend
               cmp   #ESC              ;
               beq   PrtAbort          ; Esc pressed to abort
                                       ; fall through to error counter
Seterror       inc   errcnt            ; Inc error counter
               lda   errcnt            ; 
               cmp   #$0A              ; are there 10 errors? (Xmodem spec for failure)
               bne   Resend            ; no, resend block
PrtAbort       jsr   Flush             ; yes, too many errors, flush buffer,
               jmp   Print_Err         ; print error msg and exit
Done           Jmp   Print_Good        ; All Done..Print msg and exit
;
;
; This routine will store data from the host to the SBC-3's RAM.
;
;
XModemRcv      jsr   PrintMsg          ; send prompt and info
               lda   #$01              ;
               sta   blkno             ; set block # to 1
               sta   bflag             ; set flag to get address from block 1
               lda   TargetPg          ; get start block
               sta   CurrPg            ; set current page
StartCrc       lda   #"C"              ; "C" start with CRC mode
               jsr   Put_Chr           ; send it
               lda   #$FF        
               sta   retry2            ; set loop counter for ~3 sec delay
               lda   #$00
               sta   crc
               sta   crch              ; init CRC value        
               jsr   GetByte           ; wait for input
               bcs   GotByte           ; byte received, process it
               bcc   StartCrc          ; resend "C"

StartBlk       lda   #$FF              ; 
               sta   retry2            ; set loop counter for ~3 sec delay
               jsr   GetByte           ; get first byte of block
               bcc   StartBlk          ; timed out, keep waiting...
GotByte        cmp   #ESC              ; quitting?
               bne   GotByte1          ; no
;              lda   #$FE              ; Error code in "A" of desired
               brk                     ; YES - do BRK or change to RTS if desired
GotByte1       cmp   #SOH              ; start of block?
               beq   BegBlk            ; yes
               cmp   #EOT              ;
               bne   BadCrc            ; Not SOH or EOT, so flush buffer & send NAK        
               jmp   RDone             ; EOT - all done!
BegBlk         ldx   #$00
GetBlk         lda   #$ff              ; 3 sec window to receive characters
               sta   retry2            ;
GetBlk1        jsr   GetByte           ; get next character
               bcc   BadCrc            ; chr rcv error, flush and send NAK
GetBlk2        sta   Rbuff,x           ; good char, save it in the rcv buffer
               inx                     ; inc buffer pointer        
               cpx   #$84              ; <01> <FE> <128 bytes> <CRCH> <CRCL>
               bne   GetBlk            ; get 132 characters
               ldx   #$00              ;
               lda   Rbuff,x           ; get block # from buffer
               cmp   blkno             ; compare to expected block #        
               beq   GoodBlk1          ; matched!
               jsr   Print_Err         ; Unexpected block number - abort        
               jsr   Flush             ; mismatched - flush buffer and then do BRK
;              lda   #$FD              ; put error code in "A" if desired
               brk                     ; unexpected block # - fatal error - BRK or RTS
GoodBlk1       eor   #$ff              ; 1's comp of block #
               inx                     ;
               cmp   Rbuff,x           ; compare with expected 1's comp of block #
               beq   GoodBlk2          ; matched!
               jsr   Print_Err         ; Unexpected block number - abort        
               jsr   Flush             ; mismatched - flush buffer and then do BRK
;              lda   #$FC              ; put error code in "A" if desired
               brk                     ; bad 1's comp of block#        
GoodBlk2       jsr   CalcCRC           ; calc CRC
               lda   Rbuff,y           ; get hi CRC from buffer
               cmp   crch              ; compare to calculated hi CRC
               bne   BadCrc            ; bad crc, send NAK
               iny                     ;
               lda   Rbuff,y           ; get lo CRC from buffer
               cmp   crc               ; compare to calculated lo CRC
               beq   GoodCrc           ; good CRC
BadCrc         jsr   Flush             ; flush the input port
               lda   #NAK              ;
               jsr   Put_Chr           ; send NAK to resend block
               jmp   StartBlk          ; start over, get the block again                        
GoodCrc        ldx   #$02              ;
               lda   blkno             ; get the block number
               cmp   #$01              ; 1st block?
               bne   CopyBlk           ; no, copy all 128 bytes
               lda   bflag             ; is it really block 1, not block 257, 513 etc.
               beq   CopyBlk           ; no, copy all 128 bytes
               lda   Rbuff,x           ; get target address from 1st 2 bytes of blk 1
               sta   ptr               ; save lo address
               inx                     ;
               lda   Rbuff,x           ; get hi address
               sta   ptr+1             ; save it
               inx                     ; point to first byte of data
               dec   bflag             ; set the flag so we won't get another address                
CopyBlk        ldy   #$00              ; set offset to zero
               lda   CurrPg            ; get target page var - same as Monitor Dest page
               pha                     ; put on stack
               #PLB                    ; set data page      
CopyBlk3       #LDA_ALX Rbuff, $00     ; LDA 0,Rbuff,x  abs long x addressing
               sta   (ptr),y           ; save to target
               inc   ptr               ; point to next address
               bne   CopyBlk4          ; did it step over page boundary?
               inc   ptr+1             ; adjust high address for page crossing
               bne   CopyBlk4          ; did it step over block boundary?
               inc   CurrPg            ; go to next block
               lda   CurrPg            ; get target page var - same as Monitor Dest page
               pha                     ; put on stack
               #PLB                    ; set data page      
CopyBlk4       inx                     ; point to next data byte
               cpx   #$82              ; is it the last byte
               bne   CopyBlk3          ; no, get the next one
               lda     #$00            ; set data page back to 0
               pha                     ; push to stack
               #PLB                    ; set datapage
IncBlk         inc   blkno             ; done.  Inc the block #
               lda   #ACK              ; send ACK
               jsr   Put_Chr           ;
               jmp   StartBlk          ; get next block

RDone          lda   #ACK              ; last block, send ACK and exit.
               jsr   Put_Chr           ;
               jsr   Flush             ; get leftover characters, if any
               jsr   Print_Good        ;
               rts                     ;
;
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;
; subroutines
;
Put_Chr        jmp   Ser_Out           ;

GetByte        lda   #$00              ; wait for byte from Serial port and cycle timing loop
               sta   retry             ; set low value of timing loop
StartCrcLp     jsr   Scan_SERIN        ; test if ser chr rcvd? 
               bcs   GetByte1          ; got one, so exit
               dec   retry             ; no character received, so dec counter
               bne   StartCrcLp        ;
               dec   retry2            ; dec hi byte of counter
               bne   StartCrcLp	       ; look for character again
               clc                     ; if loop times out, CLC, else SEC and return
               rts
GetByte1       jsr   SER_IN            ; get ser chr
               rts                     ; return with character in "A"
;
Flush          lda   #$70              ; flush receive buffer
               sta   retry2            ; flush until empty for ~1 sec.
Flush1         jsr   GetByte           ; read the port
               bcs   Flush             ; if chr recvd, wait for another
               rts                     ; else done
;
PrintMsg       ldx   #$00              ; PRINT starting message
PrtMsg1        lda   Msg,x                
               beq   PrtMsg2 
               pha                       
               jsr   Put_Chr
               pla
               jsr   output
               inx
               bne   PrtMsg1
PrtMsg2        rts
Msg            .byte "Begin XMODEM/CRC transfer.  Press <Esc> to abort..."
               .BYTE CR, LF
               .byte 0
;
Print_Err      ldx   #$00              ; PRINT Error message
PrtErr1        lda   ErrMsg,x
               beq   PrtErr2
               pha                       
               jsr   Put_Chr
               pla
               jsr   output
               inx
               bne   PrtErr1
PrtErr2        rts
ErrMsg         .byte "Transfer Error!"
               .BYTE CR, LF
               .byte 0
;
Print_Good     ldx   #$00              ; PRINT Good Transfer message
Prtgood1       lda   GoodMsg,x
               beq   Prtgood2
               pha                       
               jsr   Put_Chr
               pla
               jsr   output
               inx
               bne   Prtgood1
Prtgood2       rts
GoodMsg        .byte EOT,CR,LF,EOT,CR,LF,EOT,CR,LF,CR,LF
               .byte "Transfer Successful!"
               .BYTE CR, LF
               .byte 0
;
;**************************************************************************
;  CRC subroutines 
;
CalcCRC        lda   #$00              ; yes, calculate the CRC for the 128 bytes
               sta   crc               ;
               sta   crch              ;
               ldy   #$02              ;
CalcCRC1       lda   Rbuff,y           ;
               eor   crc+1             ; Quick CRC computation with lookup tables
               tax                     ; updates the two bytes at crc & crc+1
               lda   crc               ; with the byte send in the "A" register
               eor   CRCHI,X
               sta   crc+1
               lda   CRCLO,X
               sta   crc
               iny                     ;
               cpy   #$82              ; done yet?
               bne   CalcCRC1          ; no, get next
               rts                     ; y=82 on exit
;
; End of File
;
;
; Version 1.0 - First release for 65816 native mode - 11-06-2008
;