
;**********************************************************************
; 6522 - LCD module Interface Routines (c) 2001 
; By Daryl Rictor - 65c02@altavista.com
;
; Uses 163 bytes of RAM or ROM.  Requires no RAM other than up to 8 bytes
; on the system stack for subroutine calls and temporary storage.
;
; Provides software support to drive an HD44780 compatible LCD module from
; a 6522 VIA.  This is the connection diagram:
;----------------------------------------------------------------
;
;    /--------- LCD Module Connection ---------\
;     1  2  3  4  5  6  7  8  9  10 11 12 13 14  LCD pin numbers
;     o  o  o  o  o  o  o  o  o  o  o  o  o  o
;     |  |  |  |  |  |  |  |  |  |  |  |  |  |  
;     -  +  C  RS RW E  D0 D1 D2 D3 D4 D5 D6 D7  LCD Pin names
;     |  |  |  |  |  |  |  |  |  |  |  |  |  |  
;     :  :  :  :  :  :  :  :  :  :  :  :  :  :  
;     |  |  |  |  |  |  |  |  |  |  |  |  |  | 
;     |  |  | B4 B5 B6 A0 A1 A2 A3 A4 A5 A6 A7   6522 IO Port names 
;     |  |  |  |  |  |  |  |  |  |  |  |  |  | 
;     |  |  |  o  o  o  o  o  o  o  o  o  o  o 
;     |  |  | 14 15 16  2  3  4  5  6  7  8  9   6522 pin numbers
;     |  |  | \------- 6522 Connections ------/ 
;     |  |  | 
;     |  |  \---| Contrast                    
;     |  |      v                             
;     |  o---<VVVVV>-\ 10K Ohm variable resistor  
;     |  |           |                                             
;     |  o +5vdc     |                                             
;     |              |                                              
;     o--------------/                                              
;    _|_                                                       
;     -                                                       
;    gnd                                                        
;-----------------------------------------------------------------                                                            
;
; uses Via_pra for 8 bit data bus
; uses Via_prb bits 4-6 as control signals 
;     bit 4 = RS,  0=Cmd  1=Data
;     bit 5 = R/W  0=Write 1=Read
;     bit 6 = E    0=Not Enabled 1=Enabled
;
; *** All writes to PRB maintain remaining bits so other apps sharing  
; PRB are not disturbed. Use the variable definitions below to change 
; the bits that you use in your application.
;
; Library routines included:
; ---------------------------
; LCD_Byte  - Print contents of A as a 2 character HEX number
; LCD_Chr   - Print contents of A as an ASCII chracter
; LCD_Curs  - Set the Cursor Position from A (0-4F)
; LCD_Home  - Move cursor to the home position without clearing display
; LCD_Init  - Initialize the display
; LCD_Wait  - Reads LCD command register and waits for it to be idle, returns
;             with the address pointer in the A register.
; LCD_Write - Strobes the Enable Pin and waits for the LCD to process the command.
;             Uses LCD_Wait so cursor position is returned in A register.
; *** Contents of A reg are altered by these routines. Contents of X & Y registers
; are always preserved.
;
;-- Variable Definitions - (substitute your values here) --------------------------

Via_pra        =     $7f01             ; 
Via_prb        =     $7f00             ;     
Via_ddra       =     $7f03             ;     
Via_ddrb       =     $7f02             ;     
LCD_RS         =     $10               ; bit mask for the RS pin    
LCD_RW         =     $20               ; bit mask for the RW pin        
LCD_E          =     $40               ; bit mask for the E pin        
LCD_MASK       =     $8F               ; bit mask to turn off all three pins


	*=   $1000                     ; load code @ $1000

;***********************************************************************************
;
; *** LCD_Byte ***
; outputs a hex byte to LCD
; flows into LCD routine LCD_pchr()
;
LCD_Byte       pha                     ; prints AA hex digits
               LSR                     ; MOVE UPPER NIBBLE TO LOWER
               LSR                     ;
               LSR                     ;
               LSR                     ;
               JSR   LCD_byte1         ;
               pla                     ;
LCD_byte1      AND   #$0F              ; prints A hex nibble (low 4 bits)
               clc                     ;
               adc   #$30              ; convert from number to ASCII
               cmp   #$3A              ; is it HEX A-F
               bmi   LCD_chr           ; no, ready to print
               adc   #$06              ; yes, add 6 to offset gap between ASCII 9 and A

; *** LCD_Chr ***
; outputs an ascii chr to the LCD
; part of the LCD_byte routine
;
LCD_Chr        sta   Via_pra           ; Print ASCII Character to the LCD
               pha                     ; save chr
               lda   Via_prb           ;
               pha                     ;
               and   #LCD_MASK         ;
               ora   #LCD_RS           ;
               sta   Via_prb           ; write data mode
               jsr   LCD_write         ;
               pla                     ;
               sta   Via_prb           ;
               pla                     ; restore chr
               rts                     ;
;
; *** LCD_Curs ***
; Move cursor to the position in A
; This routine falls through to the LCD_Home routine. The db $2c "gobbles up" 
; the LDA #$02 command by tucking it inside a BIT instruction. Neat trick!
; Remember, cursor positions are 0-27 for line 1 and 28-4F for line 2.
;
LCD_Curs       ora   #$80              ; ddram set cmd = 1xxxxxxx
               .byte $2C               ; Mask out the LDA #$02 below and
                                       ; store incoming A reg to pra
;
; *** LCD_Home ***
; Return cursor to the home position
;
LCD_Home       lda   #$02              ;
               sta   Via_pra           ; return home cmd
               lda   Via_prb           ;
               and   #LCD_MASK         ; write cmd
               sta   Via_prb           ;
               bra   LCD_write         ; set return home cmd
;
; *** LCD_Init *** 
; Initializes the LCD with 8 data bits, 2 lines, font 5x7,
; no cursor or block cursor, clear the display, and turn it on
;
LCD_Init       lda   #$ff              ;
               sta   Via_ddra          ; set pra as all outputs
               lda   #$ff              ; built bit mask for control pins
               eor   #LCD_MASK         ; turns all three bits to 1
               ora   Via_ddrb          ; mask in with existing outputs
               sta   Via_ddrb          ; set LCD control pins as outputs      
               lda   Via_prb           ;   
               and   #LCD_MASK         ;
               sta   Via_prb           ; set LCD cont=cmd,wr,in
               lda   #$38              ;
               sta   Via_pra           ; set funct=8,2,0
               jsr   LCD_write         ;
               lda   #$01              ;
               sta   Via_pra           ; clear disp
               jsr   LCD_write         ;
               lda   #$0e              ; 0c
               sta   Via_pra           ; set disp on, no cursor, no block
               jsr   LCD_write         ;
               rts                     ; done
;
; *** LCD_Write ***
; toggles the LCD enable pin of the 6522
; Falls into LCD_Wait()
;
;
LCD_write      lda   Via_prb           ;
               ora   #LCD_E            ; raise enable pin
               sta   Via_prb           ;
               lda   Via_prb           ; waste some time...
               eor   #LCD_E            ; to allow for enable timing
               sta   Via_prb           ; lower enable pin
;
; *** LCD_Wait ***
; Scans the LCD busy line (reads cmd reg bit 7) and loops until its idle (0).
; Upon return, A reg contains the Cursor position
;
LCD_Wait       lda   #$00              ; 
               sta   Via_ddra          ; set pra for input
               lda   Via_prb           ;
               phx                     ; save x register temporarily
               pha                     ; save prb state
               and   #LCD_MASK         ; 
               ora   #LCD_RW           ;
               ora   #LCD_E            ;
               sta   Via_prb           ; set cmd read mode
LCD_wait1      lda   Via_pra           ;
               bmi   LCD_wait1         ; loop unitl bit 7=0
               tax                     ; save pra contents into X (cursor pos)
               pla                     ; restore prb state
               sta   Via_prb           ; 
               lda   #$ff              ; set pra for output
               sta   Via_ddra          ;
               txa                     ; put pra contents in A (cursor pos)
               plx                     ; restore x register
               rts                     ; 
;
; end of LCD Routines
;*************************************************************************