;
; Graphics driver for the 320x200x256 mode on SBC-3.
;
; Based on Stephen L. Judds GRLIB code
;
;
; Modified for SBC-3 by Daryl Rictor
;
	.include	"zeropage.inc"
	.include	"sbc3.inc"
	.include	"tgi-kernel.inc"
	.include	"tgi-mode.inc"
	.include	"tgi-error.inc"

	.export		_sbc3_tgi
	.macpack	generic


; ------------------------------------------------------------------------
; Header. Includes jump table and constants.

.segment  "CODE"

_sbc3_tgi:
; First part of the header is a structure that has a magic and defines the
; capabilities of the driver

	.byte	$74, $67, $69		; "tgi"
	.byte	TGI_API_VERSION		; TGI API version number
xres:	.word	320			; X resolution
yres:	.word	200			; Y resolution
	.byte	255			; Number of drawing colors
	.byte	2			; Number of screens available
	.byte	8			; System font X size
	.byte	8			; System font Y size
	.res	4, $00			; Reserved for future extensions

; Next comes the jump table. Currently all entries must be valid and may point
; to an RTS for test versions (function not implemented). A future version may
; allow for emulation: In this case the vector will be zero. Emulation means
; that the graphics kernel will emulate the function by using lower level
; primitives - for example ploting a line by using calls to SETPIXEL.

	.addr	INSTALL
	.addr	UNINSTALL
	.addr	INIT
	.addr	DONE
	.addr	GETERROR
	.addr	CONTROL
	.addr	CLEAR
	.addr	SETVIEWPAGE
	.addr	SETDRAWPAGE
	.addr	SETCOLOR
	.addr	SETPALETTE
	.addr	GETPALETTE
	.addr	GETDEFPALETTE
	.addr	SETPIXEL
	.addr	GETPIXEL
	.addr	LINE
	.addr	BAR
	.addr	CIRCLE
	.addr	TEXTSTYLE
	.addr	OUTTEXT
	.addr	0			; IRQ entry is unused

; ------------------------------------------------------------------------
; Data.

; Variables mapped to the zero page segment variables. Some of these are
; used for passing parameters to the driver.

X1	:= ptr1
Y1	:= ptr2
XP	:= ptr1				; circle
YP	:= ptr2				; circle
X2	:= ptr3
Y2	:= ptr4
RADIUS	:= tmp1

ROW	:= tmp2				; Bitmap row...
COL	:= tmp3				; ...and column, both set by PLOT
TEMP	:= tmp4
TEMP2	:= sreg
POINT	:= $8A
INRANGE	:= regsave+2			; PLOT variable, $00 = coordinates in range

XX	:= ptr3				; circle
YY	:= ptr4				; circle

; Absolute variables used in the code

.bss

ERROR:	.res	1			; Error code
PALETTE: .res	2			; The current palette

BITMASK: .res    1			; $00 = clear, $FF = set pixels
DPAGE:	.res	1			; draw page 

; Line routine stuff

FX:						; circle
DX:	.res	2
FY:						; circle
DY:	.res	1
IX:	.res	1
IY:	.res	1

; Circle routine stuff, overlaid by BAR variables

FF:
PP:
X1SAVE:	.res	2
XC:
Y1SAVE:	.res	2
YC:
X2SAVE:	.res	2
OF:
Y2SAVE:	.res	2

; Text output stuff
TEXTMAGX:  .res	1
TEXTMAGY:  .res	1
TEXTDIR:  .res	1

; Constants and tables

.rodata

DEFPALETTE:  .byte   $00, $FF		; White on black
PALETTESIZE  = * - DEFPALETTE

.code

; ------------------------------------------------------------------------
; INSTALL routine. Is called after the driver is loaded into memory. May
; initialize anything that has to be done just once. Is probably empty
; most of the time.
;
; Must set an error code: NO
;

INSTALL:
	rts


; ------------------------------------------------------------------------
; UNINSTALL routine. Is called before the driver is removed from memory. May
; clean up anything done by INSTALL but is probably empty most of the time.
;
; Must set an error code: NO
;

UNINSTALL:
	rts


; ------------------------------------------------------------------------
; INIT: Changes an already installed device from text mode to graphics
; mode.
; Note that INIT/DONE may be called multiple times while the driver
; is loaded, while INSTALL is only called once, so any code that is needed
; to initializes variables and so on must go here. Setting palette and
; clearing the screen is not needed because this is called by the graphics
; kernel later.
; The graphics kernel will never call INIT when a graphics mode is already
; active, so there is no need to protect against that.
;
; Must set an error code: YES
;

INIT:

; Initialize variables

	lda	#$00
	jsl	_scursor		; turn cursor off        
	ldx	#$FF
	stx	BITMASK
	lda	#$01
	sta	DPAGE			; bank 1 on sbc3
	lda	#TGI_ERR_OK
	sta	ERROR
	rts

; ------------------------------------------------------------------------
; DONE: Will be called to switch the graphics device back into text mode.
; The graphics kernel will never call DONE when no graphics mode is active,
; so there is no need to protect against that.
;
; Must set an error code: NO
;

DONE:
	lda	#'_'
	jsl	_scursor		; turn cursor on
	lda	#$01
	sta	$07FFFF			; set display back to default
	dec
	jsr	SETVIEWPAGE 
	lda	#$FF
	ldx	#$00
	jsl	_setc			; set colors
	rts

; ------------------------------------------------------------------------
; GETERROR: Return the error code in A and clear it.

GETERROR:
     	ldx	#TGI_ERR_OK
	lda	ERROR
	stx	ERROR
        rts

; ------------------------------------------------------------------------
; CONTROL: Platform/driver specific entry point.
;
; Must set an error code: YES
;

CONTROL:
	lda	#TGI_ERR_INV_FUNC
	sta	ERROR
        rts

; ------------------------------------------------------------------------
; CLEAR: Clears the screen.
;
; Must set an error code: NO
;

CLEAR:
	  lda	#$ff
	  ldx #0
	  jsl _setc	
	  jsl	_cls	
        rts

; ------------------------------------------------------------------------
; SETVIEWPAGE: Set the visible page. Called with the new page in A (0..n).
; The page number is already checked to be valid by the graphics kernel.
;
; Must set an error code: NO (will only be called if page ok)
;
; valid range is 0 or 1
SETVIEWPAGE:
        inc 
        and	#$07	; mask valid page range
        sta	temp	; save
	  lda	$080030
	  and #$F8	; mask page range out
	  ora temp
	  sta $080030
        rts

; ------------------------------------------------------------------------
; SETDRAWPAGE: Set the drawable page. Called with the new page in A (0..n).
; The page number is already checked to be valid by the graphics kernel.
;
; Must set an error code: NO (will only be called if page ok)
;

SETDRAWPAGE:
        inc 
        sta DPAGE
	  sta $07FFFF
        rts

; ------------------------------------------------------------------------
; SETCOLOR: Set the drawing color (in A). The new color is already checked
; to be in a valid range (0..maxcolor-1).
;
; Must set an error code: NO (will only be called if color ok)
;

SETCOLOR:
        sta     BITMASK
        rts

; ------------------------------------------------------------------------
; SETPALETTE: Set the palette (not available with all drivers/hardware).
; A pointer to the palette is passed in ptr1. Must set an error if palettes
; are not supported
;
; Must set an error code: YES
;

SETPALETTE:
	lda	#TGI_ERR_INV_FUNC
      sta     ERROR
      rts

; ------------------------------------------------------------------------
; GETPALETTE: Return the current palette in A/X. Even drivers that cannot
; set the palette should return the default palette here, so there's no
; way for this function to fail.
;
; Must set an error code: NO
;

GETPALETTE:
; fall thru to GETDEFPALETTE	

; ------------------------------------------------------------------------
; GETDEFPALETTE: Return the default palette for the driver in A/X. All
; drivers should return something reasonable here, even drivers that don't
; support palettes, otherwise the caller has no way to determine the colors
; of the (not changeable) palette.
;
; Must set an error code: NO (all drivers must have a default palette)
;

GETDEFPALETTE:
        lda     #<DEFPALETTE
        ldx     #>DEFPALETTE
        rts

; ------------------------------------------------------------------------
; SETPIXEL: Draw one pixel at X1/Y1 = ptr1/ptr2 with the current drawing
; color. The coordinates passed to this function are never outside the
; visible screen area, so there is no need for clipping inside this function.
;
; Must set an error code: NO
;

SETPIXEL:
        jsr     CALC            ; Calculate coordinates
        lda     BITMASK
        sta     [POINT]
        rts

SETPIXELCLIP:
      lda	Y1+1
	bmi	:+		; y < 0
	lda	X1+1
	bmi	:+		; x < 0
	lda	X1
	ldx	X1+1
      sec
      sbc	xres
      txa
	sbc	xres+1
	bcs	:+
	lda	Y1
	ldx	Y1+1
      sec
      sbc	yres
	txa
      sbc	yres+1
	bcc	SETPIXEL
:	rts

; ------------------------------------------------------------------------
; GETPIXEL: Read the color value of a pixel and return it in A/X. The
; coordinates passed to this function are never outside the visible screen
; area, so there is no need for clipping inside this function.


GETPIXEL:
        jsr     CALC            ; Calculate coordinates
        lda     [POINT]
        ldx     #$00            ; Clear high byte
        rts

; ------------------------------------------------------------------------
; LINE: Draw a line from X1/Y1 to X2/Y2, where X1/Y1 = ptr1/ptr2 and
; X2/Y2 = ptr3/ptr4 using the current drawing color.
;
; Must set an error code: NO
;

LINE:

@CHECK: lda     X2           ;Make sure x1<x2
        sec
        sbc     X1
        tax
        lda     X2+1
        sbc     X1+1
        bpl     @CONT
        lda     Y2           ;If not, swap P1 and P2
        ldy     Y1
        sta     Y1
        sty     Y2
        lda     Y2+1
        ldy     Y1+1
        sta     Y1+1
        sty     Y2+1
        lda     X1
        ldy     X2
        sty     X1
        sta     X2
        lda     X2+1
        ldy     X1+1
        sta     X1+1
        sty     X2+1
        bcc     @CHECK

@CONT:  lsr
	  sta     DX+1
        txa			
        ror		
        sta     DX
        bne     @NVT  
        jmp     @VERT
@NVT:   stz     IX
        stz     IY	
        sec
        lda     Y2           ;Calculate dy
        sbc     Y1
        bne     @NHZ
	  jmp     @HORZ
@NHZ:	  bcc	    @YNEG
@YPOS:  lsr
        sta     DY
        jsr    SETPIXELCLIP		;plot first point
@LOOP0: ldx     #$00
@LOOP:  clc
        lda     IX
        adc     DX
        sta     IX
        bcc     @LOOP1
        inx
        inc     X1
        bne     @LOOP1
        inc     X1+1
@LOOP1: clc
        lda     IY
        adc     DY
        sta     IY
        bcc     @LOOP2
	  inx
        INC     Y1  
@LOOP2: txa
        beq     @LOOP     ; no adjustment yet
        jsr    SETPIXELCLIP
        lda    DX
        beq    @END0      ; if DX=0, skip x end test
        lda     X1
        cmp     X2
        bne     @END0
        lda     X1+1 
        cmp     X2+1
        beq     @END
@END0:  lda     DY
        beq     @LOOP0    ; if dy=0, skip y end test   
        lda     Y1
        cmp     Y2
        bne     @LOOP0    ; reset x=0         
@END:   lda     X2
        sta     X1
        lda     X2+1
        sta     X1+1
        lda     Y2
        sta     Y1
        jsr     SETPIXELCLIP
	  rts	

@YNEG:  jsr     SETPIXELCLIP		;plot first point
        lda     Y1
        sec
        sbc     Y2
        lsr
        sta     DY
@LOOPX: ldx     #$00
@LOOP3: clc
        lda     IX
        adc     DX
        sta     IX
        bcc     @LOOP4
        inx
        inc     X1
        bne     @LOOP4
        inc     X1+1
@LOOP4: clc
        lda     IY
        adc     DY
        sta     IY
        bcc     @LOOP5
	  inx
        DEC     Y1  
@LOOP5: txa
        beq     @LOOP3     ; no adjustment yet
        jsr    SETPIXELCLIP
        lda     DX
        beq     @LOOP6      ; if DX=0, skip x end test
        lda     X1
        cmp     X2
        bne     @LOOP6
        lda     X1+1
        cmp     X2+1
        beq     @END
@LOOP6: lda     DY
        beq     @LOOPX      ; reset x
        lda     Y1
        cmp     Y2
        beq     @END
        bra     @LOOPX         

@VERT:  sec
        lda     Y2           ;Calculate dy
        sbc     Y1
	  bcc	    @VNEG
        bra     @VP2
@VP1:   inc    Y1
@VP2:   jsr    SETPIXELCLIP
        sec
	  lda    Y2
        sbc    Y1
        bne    @VP1
        rts
@VN1:   dec    Y1
@VNEG:  jsr    SETPIXELCLIP
        sec
	  lda    Y2
        sbc    Y1
        bne    @VN1
        rts

@HRZ:   inc    X1
        bne    @HORZ
        inc    X1+1
@HORZ:  jsr    SETPIXELCLIP
        sec
	  lda    X2
        sbc    X1
        bne    @HRZ
        lda    X2+1
        sbc    X1+1
        bne    @HRZ
        rts

; ------------------------------------------------------------------------
; BAR: Draw a filled rectangle with the corners X1/Y1, X2/Y2, where
; X1/Y1 = ptr1/ptr2 and X2/Y2 = ptr3/ptr4 using the current drawing color.
; Contrary to most other functions, the graphics kernel will sort and clip
; the coordinates before calling the driver, so on entry the following
; conditions are valid:
;       X1 <= X2
;       Y1 <= Y2
;       (X1 >= 0) && (X1 < XRES)
;       (X2 >= 0) && (X2 < XRES)
;       (Y1 >= 0) && (Y1 < YRES)
;       (Y2 >= 0) && (Y2 < YRES)
;
; Must set an error code: NO
;

; Note: This function needs optimization. It's just a cheap translation of
; the original C wrapper and could be written much smaller (besides that,
; calling LINE is not a good idea either).

BAR:	  lda     Y2
        sta     Y2SAVE
        lda     Y2+1
        sta     Y2SAVE+1

        lda     X2
        sta     X2SAVE
        lda     X2+1
        sta     X2SAVE+1

        lda     Y1
        sta     Y1SAVE
        lda     Y1+1
        sta     Y1SAVE+1

        lda     X1
        sta     X1SAVE
        lda     X1+1
        sta     X1SAVE+1

@L1:	  lda     Y1
        sta     Y2
        lda     Y1+1
        sta     Y2+1
  	  jsr     LINE

        lda     Y1SAVE
        cmp     Y2SAVE
        bne     @L2
        lda     Y1SAVE
        cmp     Y2SAVE
        beq     @L4

@L2:    inc     Y1SAVE
        bne     @L3
        inc     Y1SAVE+1

@L3:    lda     Y1SAVE
        sta     Y1
        lda     Y1SAVE+1
        sta     Y1+1

        lda     X1SAVE
        sta     X1
        lda     X1SAVE+1
        sta     X1+1

        lda     X2SAVE
        sta     X2
        lda     X2SAVE+1
        sta     X2+1
        jmp     @L1

@L4:    rts


; ------------------------------------------------------------------------
; CIRCLE: Draw a circle around the center X1/Y1 (= ptr1/ptr2) with the
; radius in tmp1 and the current drawing color.
;
; Must set an error code: NO
;
CIRCLE: 
	lda	RADIUS
	bne	:+
	jmp	SETPIXELCLIP	; Plot as a point and exit

;  int y = radius;
:	sta	YY
	stz	YY+1

;  int x = 0;
	stz	XX
	stz	XX+1

; save center point
	lda	XP
	sta	XC
	lda	XP+1
	sta	XC+1
	lda	YP
	sta	YC
	lda	YP+1
	sta	YC+1

; int f = 1 - radius;
	sec
	lda	#$01
	sbc	RADIUS
	sta	FF
	stz	FF+1
	bcs   :+
	dec	FF+1

; int ddF_x = 1;
: 	lda	#$01
	sta	FX
	stz	FX+1

; int ddF_y = -2 * radius;
	stz	FY+1
	lda	RADIUS
	asl			; *2
	sta	FY
	rol	FY+1
	lda   FY
	EOR	#$FF
      sta	FY
	lda	FY+1
	EOR	#$FF
	sta	FY+1
	inc	FY
	bne	:+
	inc	FY+1

;  tgi_setpixel(xC, yC + y);
:	lda	XC
	sta	XP
	lda	XC+1
	sta	XP+1
	clc
	lda	YC
	adc	YY
	sta	YP
	lda	YC+1
	adc	YY+1
	sta	YP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC, yC - y);
	sec
	lda	YC
	sbc	YY
	sta	YP
	lda	YC+1
	sbc	YY+1
	sta	YP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC + y, yC);
	clc
	lda	XC
	adc	YY
	sta	XP
	lda	XC+1
	adc	YY+1
	sta	XP+1
	lda	YC
	sta	YP
	lda	YC+1
	sta	YP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC - y, yC);
	sec
	lda	XC
	sbc	YY
	sta	XP
	lda	XC+1
	sbc	YY+1
	sta	XP+1
	jsr	SETPIXELCLIP

@LOOP:
;  while (x < y) {
	sec
	lda	XX
	sbc	YY
	lda	XX+1
	sbc	YY+1
	bcc	@1	; x<y
	rts

@1:	lda	FF+1
	bmi	@2

	lda	YY
	bne	@11
	dec	YY+1
@11:	dec	YY
	clc
	lda	FY
	adc	#$02
	sta	FY
	tax
	lda	FY+1
	adc	#$00
	sta	FY+1
	tay
	clc
	txa
	adc	FF
	sta	FF
	tya
	ADC	FF+1
	sta	FF+1

@2:	inc	XX
	bne	@21
	inc	XX+1
@21:	clc
	lda	FX
	adc	#$02
	sta	FX
	tax
	lda	FX+1
	adc	#$00
	sta	FX+1
	tay
	clc
	txa
	adc	FF
	sta	FF
	tya
	ADC	FF+1
	sta	FF+1

;  tgi_setpixel(xC + x, yC + y);
	clc
	lda	XC
	adc	XX
	sta	XP
	pha
	lda	XC+1
	adc	XX+1
	sta	XP+1
	pha
	clc
	lda	YC
	adc	YY
	sta	YP
	lda	YC+1
	adc	YY+1
	sta	YP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC - x, yC + y);
	sec
	lda	XC
	sbc	XX
	sta	XP
	lda	XC+1
	sbc	XX+1
	sta	XP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC - x, yC - y);
	sec
	lda	YC
	sbc	YY
	sta	YP
	lda	YC+1
	sbc	YY+1
	sta	YP+1
	jsr	SETPIXELCLIP


;  tgi_setpixel(xC + x, yC - y);
	pla
	sta	XP+1
	pla	
	sta	XP
	jsr	SETPIXELCLIP



;  tgi_setpixel(xC + y, yC + x);
	clc
	lda	XC
	adc	YY
	sta	XP
	pha
	lda	XC+1
	adc	YY+1
	sta	XP+1
	pha
	clc
	lda	YC
	adc	XX
	sta	YP
	lda	YC+1
	adc	XX+1
	sta	YP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC - y, yC + x);
	sec
	lda	XC
	sbc	YY
	sta	XP
	lda	XC+1
	sbc	YY+1
	sta	XP+1
	jsr	SETPIXELCLIP

;  tgi_setpixel(xC - y, yC - x);
	sec
	lda	YC
	sbc	XX
	sta	YP
	lda	YC+1
	sbc	XX+1
	sta	YP+1
	jsr	SETPIXELCLIP


;  tgi_setpixel(xC + x, yC - y);
	pla
	sta	XP+1
	pla	
	sta	XP
	jsr	SETPIXELCLIP
	jmp	@LOOP


; ------------------------------------------------------------------------
; TEXTSTYLE: Set the style used when calling OUTTEXT. Text scaling in X and Y
; direction is passend in X/Y, the text direction is passed in A.
;
; Must set an error code: NO
;

TEXTSTYLE:
        stx     TEXTMAGX
        sty     TEXTMAGY
        sta     TEXTDIR
        rts


; ------------------------------------------------------------------------
; OUTTEXT: Output text at X/Y = ptr1/ptr2 using the current color and the
; current text style. The text to output is given as a zero terminated
; string with address in ptr3.
;
; Must set an error code: NO
;

OUTTEXT:
        jsr CALC       ; set ROW,COL
        lda ROW
	  jsl _sety
        lda COL
        jsl _setx
        lda BITMASK
        ldx #0         ; bg black
        jsl _setc
        ldy #0
@1:     lda (ptr3),y
        beq @2
        jsl _cout
        iny
        bne @1 
@2:     rts

; ------------------------------------------------------------------------
; Calculate all variables to plot the pixel at X1/Y1. If the point is out
; of range, a carry is returned and INRANGE is set to a value !0 zero. If
; the coordinates are valid, INRANGE is zero and the carry clear.

CALC:   lda     Y1
        sta     ROW
        lda     Y1+1
        lsr                     ; Neg is possible
        ror     ROW
        lsr
        ror     ROW
        lsr
        ror     ROW

        lda     #00
        sta     POINT
        lda     Y1
        lsr
        ror     POINT
        ror
        ror     POINT           ; row*64
        adc     Y1             ; +row*256
        sta     POINT+1

        lda     X1
        sta     COL
        lda     X1+1
        lsr
        ror     COL
        lsr
        ror     COL
        lsr
        ror     COL

        clc
        lda     X1  
        adc     POINT           ; 
        sta     POINT
        lda     X1+1
        adc     POINT+1
        sta     POINT+1
	  lda     DPAGE		  ; get drawing page
	  sta	    POINT+2

        lda     ROW
        cmp     #25
        bcs     @ERR
        lda     COL
        cmp     #40
        bcs     @ERR
	  clc
        lda     #00
        sta     INRANGE
        rts

@ERR:    sec
	  lda     #255
        sta     INRANGE
        rts




