;
; Serial driver 
;
; 
;
; Modified for SBC-3 by Daryl Rictor
;

	.include	"zeropage.inc"
	.include	"ser-kernel.inc"
	.include	"ser-error.inc"
	.include	"sbc3.inc"

	.export		_sbc3_ser

; ------------------------------------------------------------------------
; Header. Includes jump table

.code

; Driver signature

_sbc3_ser:
	.byte	$73, $65, $72		; "ser"
	.byte	SER_API_VERSION		; Serial API version number

; Jump table.

	.word	INSTALL
	.word	UNINSTALL
	.word	OPEN
	.word	CLOSE
	.word	GET
	.word	PUT
	.word	STATUS
	.word	IOCTL
	.word	IRQ

;----------------------------------------------------------------------------
;
; Global variables
;

.rodata

; Tables used to translate RS232 params into register values

BaudTable:			; bit7 = 1 means setting is invalid

	.byte	$FF		; SER_BAUD_45_5
	.byte	$FF		; SER_BAUD_50
	.byte	$FF		; SER_BAUD_75
	.byte	$FF		; SER_BAUD_110
	.byte	$FF		; SER_BAUD_134_5
	.byte	$FF		; SER_BAUD_150
	.byte	$FF		; SER_BAUD_300
	.byte	$FF		; SER_BAUD_600
	.byte	$FF		; SER_BAUD_1200
	.byte	$FF		; SER_BAUD_1800
	.byte	$FF		; SER_BAUD_2400
	.byte	$FF		; SER_BAUD_3600
	.byte	$00		; SER_BAUD_4800
	.byte	$FF		; SER_BAUD_7200
	.byte	$01		; SER_BAUD_9600
	.byte	$02		; SER_BAUD_19200
	.byte	$03		; SER_BAUD_38400
	.byte	$04		; SER_BAUD_57600
        .byte   $05		; SER_BAUD_115200
        .byte   $06		; SER_BAUD_230400
        .byte   $FF		; SER_BAUD_31250
        .byte   $FF		; SER_BAUD_62500

BitTable:
        .byte   $FF		; SER_BITS_5
        .byte   $FF		; SER_BITS_6
        .byte   $00		; SER_BITS_7
        .byte   $08		; SER_BITS_8

StopTable:
        .byte   $00		; SER_STOP_1
        .byte   $40		; SER_STOP_2

ParityTable:
        .byte   $00		; SER_PAR_NONE
        .byte   $30		; SER_PAR_ODD
        .byte   $20		; SER_PAR_EVEN
        .byte   $FF		; SER_PAR_MARK
        .byte   $FF		; SER_PAR_SPACE

HandshakeTable:
        .byte   $00		; SER_HS_NONE
        .byte   $FF		; SER_HS_HW
        .byte   $FF		; SER_HS_SW


.code

;----------------------------------------------------------------------------
; INSTALL routine. Is called after the driver is loaded into memory. If
; possible, check if the hardware is present.
; Must return an SER_ERR_xx code in a/x.
;
; Since we don't have to manage the IRQ vector on the Plus/4, this is actually
; the same as:
;
; UNINSTALL routine. Is called before the driver is removed from memory.
; Must return an SER_ERR_xx code in a/x.
;
; and:
;
; CLOSE: Close the port, disable interrupts and flush the buffer. Called
; without parameters. Must return an error code in a/x.
;

INSTALL:
UNINSTALL:
CLOSE:
	lda	#<SER_ERR_OK	; Done, return an error code
	tax			; A is zero
	rts

;----------------------------------------------------------------------------
; PARAMS routine. A pointer to a ser_params structure is passed in ptr1.
; Must return an SER_ERR_xx code in a/x.

OPEN:

; Check if the handshake setting is valid

	ldy	#SER_PARAMS::HANDSHAKE	; Handshake
	lda	(ptr1),y
	cmp	#SER_HS_NONE		; This is all we support
	bne	InvParam

; Set the value for the control register, which contains stop bits, word
; length and the baud rate.

	ldy	#SER_PARAMS::BAUDRATE
	lda	(ptr1),y		; Baudrate index
	tay
	lda	BaudTable,y		; Get 6551 value
	bmi	InvBaud			; Branch if rate not supported
	sta	tmp1

	ldy	#SER_PARAMS::DATABITS	; Databits
	lda	(ptr1),y
	tay
	lda	BitTable,y
	bmi	InvParam
	ora	tmp1
	sta	tmp1

	ldy	#SER_PARAMS::STOPBITS	; Stopbits
	lda	(ptr1),y
	tay
	lda	StopTable,y
	ora	tmp1
	jsl	_setser			; sbc3 routine

	lda	#<SER_ERR_OK
	tax				; A is zero
	rts

; Invalid parameter

InvParam:
	lda	#<SER_ERR_INIT_FAILED
	ldx	#>SER_ERR_INIT_FAILED
	rts

; Baud rate not available

InvBaud:
	lda	#<SER_ERR_BAUD_UNAVAIL
	ldx	#>SER_ERR_BAUD_UNAVAIL
	rts

;----------------------------------------------------------------------------
; GET: Will fetch a character from the receive buffer and store it into the
; variable pointer to by ptr1. If no data is available, SER_ERR_NO_DATA is
; return.
;

GET:
	jsl	_scanser
	bcs	@get1
	lda	#<SER_ERR_NO_DATA
	ldx	#>SER_ERR_NO_DATA
	rts
@get1:
	jsl	_serin
	ldx	#0
	rts

;----------------------------------------------------------------------------
; PUT: Output character in A.
; Must return an error code in a/x.
;

PUT:
	jsl	_serout
	lda	#<SER_ERR_OK
	tax				; A is zero
	rts

;----------------------------------------------------------------------------
; STATUS: Return the status in the variable pointed to by ptr1.
; Must return an error code in a/x.
;

STATUS:	lda	#0
	sta	(ptr1)
	tax				; SER_ERR_OK
	rts

;----------------------------------------------------------------------------
; IOCTL: Driver defined entry point. The wrapper will pass a pointer to ioctl
; specific data in ptr1, and the ioctl code in A.
; Must return an error code in a/x.
;

IOCTL:	lda	#<SER_ERR_INV_IOCTL	; We don't support ioclts for now
	ldx	#>SER_ERR_INV_IOCTL
	rts

;----------------------------------------------------------------------------
; IRQ: Called from the builtin runtime IRQ handler as a subroutine. All
; registers are already save, no parameters are passed, but the carry flag
; is clear on entry. The routine must return with carry set if the interrupt
; was handled, otherwise with carry clear.
;

IRQ:	rts				; not used

;
; end
;